from ..abstract_optimizer import AbstractOptimizer

from jax.tree_util import tree_map


class Wrap(AbstractOptimizer):
    r"""Wrapper for Jax optimizers."""

    def __init__(self, machine, optimizer, repr_str=None):
        r"""
        Constructs a new ``Jax`` optimizer that can be used in NetKet drivers.

        Args:
            machine (AbstractMachine): The machine to be optimized.
            optimizer (jax.experimental.optimizers): A Jax optimizer.

        Examples:
            Simple SGD optimizer from Jax.

            >>> from netket.optimizer import Jax
            >>> from jax.experimental.optimizers import sgd as Sgd
            >>>

        """
        self._init_fun, self._update_fun, self._params_fun = optimizer

        self._machine = machine

        self._repr_str = repr_str

        self.reset()

    def update(self, grad, pars):
        self._opt_state = self._update_fun(self._step, grad, self._opt_state)
        self._step += 1
        pars = self._params_fun(self._opt_state)

        return pars

    def reset(self):
        self._step = 0
        self._opt_state = self._init_fun(self._machine.parameters)

    def __repr__(self):
        repr_str = self._repr_str
        if repr_str is None:
            try:
                module = self._update_fun.__module__
                qname = self._update_fun.__qualname__
                repr_str = module + "." + qname

                if module.startswith("jax"):
                    repr_str = repr_str.rsplit(".", 2)[0]
            except:
                repr_str = ""

            rep = "Wrapper for the following Jax optimizer :\n  "
            rep += str(repr_str)
            return rep



from .metropolis_hastings import MetropolisHastings
from .metropolis_hastings_pt import MetropolisHastingsPt

from .local_kernel import _LocalKernel
from .exchange_kernel import _ExchangeKernel
from .custom_kernel import _CustomKernel
from .hamiltonian_kernel import _HamiltonianKernel



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import numpy as _np
import scipy.integrate as _scint
from tqdm import tqdm

import netket.legacy as _nk
from netket.logging import JsonLog as _JsonLog
from jax.tree_util import tree_map as _tree_map

from netket.utils.mpi import (
    n_nodes as _n_nodes,
    node_number as _rank,
)

_NaN = float("NaN")


def _make_op(op, matrix_type):
    if matrix_type == "sparse":
        return op.to_sparse()
    elif matrix_type == "dense":
        return op.to_dense()
    elif matrix_type == "direct":
        return op.to_linear_op()


def _make_rhs(hamiltonian, propagation_type):
    if propagation_type == "real":

        def rhs(t, state):
            ham = hamiltonian(t)
            return -1j * ham.dot(state)

        return rhs

    elif propagation_type == "imaginary":

        def rhs(t, state):
            ham = hamiltonian(t)
            v0 = ham.dot(state)
            mean = _np.vdot(state, v0)
            return -v0 + mean * state

        return rhs


class _MockMachine:
    """
    A hack to make JsonLog work with full state data.
    """

    def __init__(self, state):
        self._state = state

    def save(self, filename):
        with open(filename, "w") as f:
            json.dump({"StateVector": [[z.real, z.imag] for z in self._state]}, f)


class PyExactTimePropagation:
    r"""
    Solver for exact real and imaginary time evolution, wrapping `scipy.integrate`
    for direct use within NetKet.
    """

    def __init__(
        self,
        hamiltonian,
        initial_state,
        dt,
        t0=0.0,
        propagation_type="real",
        matrix_type="sparse",
        solver=None,
        solver_kwargs={},
    ):
        r"""
        Create the exact time evolution driver.

           Args:
               :hamiltonian: Hamiltonian of the system.
               :initial_state: Initial state at t0.
               :dt (float): Simulation time step.
               :t0 (float): Initial time.
               :propagation_type: Specifies whether the imaginary or real-time
                    Schroedinger equation is solved. Should be one of "real" or
                    "imaginary".
               :matrix_type: The type of matrix used for the Hamiltonian when
                    creating the matrix wrapper. The default is `sparse`. The
                    other choices are `dense` and `direct`.
                :solver: A solver class, which should be a subclass of `scipy.integrate.OdeSolver`.
                    The default solver is `scipy.integrate.RK45`.
                :solver_kwargs (dict): Keyword arguments that are passed to the solver class,
                    e.g., for `atol` and `rtol`.

           Examples:
               Solving 1D Ising model with imaginary time propagation:

               >>> import netket as nk
               >>> import numpy as np
               >>> L = 8
               >>> graph = nk.graph.Hypercube(L, n_dim=1, pbc=True)
               >>> hi = nk.hilbert.Spin(1/2, N=graph.n_nodes)
               >>> hamiltonian = nk.operator.Ising(hi, h=1.0, graph=graph)
               >>> psi0 = np.ones(hi.n_states)
               >>> driver = nk.exact.PyExactTimePropagation(hamiltonian, t0=0, dt=0.1,
               ...                                          initial_state=psi0,
               ...                                          propagation_type="imaginary")
               >>> for step in driver.iter(n_iter=3):
               ...     print(driver.estimate(hamiltonian))
               -2.048e+03 ± 0.000e+00 [σ²=-4.176e+06]
               -1.107e+01 ± 0.000e+00 [σ²=-1.756e+01]
               -9.860e+00 ± 0.000e+00 [σ²=-9.977e-01]
        """
        if isinstance(hamiltonian, _nk.operator.AbstractOperator):
            self._h_op = _make_op(hamiltonian, matrix_type)
            self._h = lambda t: self._h_op
        else:
            raise NotImplementedError("Time-dependent hHamiltonian not yet supported.")
        self._rhs = _make_rhs(self._h, propagation_type)

        self._mynode = _rank
        self._mpi_nodes = _n_nodes

        self.state = initial_state
        self._dt = dt
        self._t0 = t0
        self._t = t0
        self._steps = 0
        self._prop_type = propagation_type
        self._matrix_type = matrix_type

        Solver = solver if solver else _scint.RK45

        self._solver = Solver(
            self._rhs,
            self._t0,
            self.state,
            t_bound=_np.inf,
            max_step=dt,
            **solver_kwargs,
        )

    def _estimate_stats(self, obs):
        if isinstance(obs, _nk.operator.AbstractOperator):
            op = _make_op(obs, self._matrix_type)
        else:
            op = obs

        v0 = op.dot(self.state)
        mean = _np.vdot(self.state, v0)
        variance = _np.vdot(self.state, op.dot(v0)) - mean ** 2

        return _nk.stats.Stats(mean=mean, error_of_mean=0.0, variance=variance.real)

    def estimate(self, observables):
        return _tree_map(self._estimate_stats, observables)

    @property
    def t(self):
        r"""
        Current simulation time.
        """
        return self._t

    @property
    def dt(self):
        r"""
        Simulation time step.
        """
        return self._dt

    @property
    def step_count(self):
        return self._step

    def advance(self, n_steps):
        r"""
        Advance the time propagation by `n_steps` simulation steps
        of duration `self.dt`.

           Args:
               :n_steps (int): No. of steps to advance.
        """
        t_end = self._t + n_steps * self._dt

        self._solver.t = self.t
        self._solver.t_bound = t_end
        self._solver.status = "running"
        while self._solver.status == "running":
            self._solver.step()
        if self._solver.status == "failed":
            raise ...

        self.state = self._solver.y
        self._t = self._solver.t

    def iter(self, n_iter, step=1):
        """
        Returns a generator which advances the time evolution in
        steps of `step` for a total of `n_iter` times.

        Args:
            :n_iter (int): The total number of steps.
            :step (int=1): The size of each step.

        Yields:
            :(int): The current step.
        """
        for i in range(n_iter):
            # First yield, then step for compatibility with AbstractVariationalDriver
            yield i
            self.advance(step)

    def reset(self):
        """
        Resets the driver.
        """
        self.t0 = self.t
        self._step = 0

    def run(
        self,
        output_prefix,
        n_iter,
        obs=None,
        save_params_every=50,
        write_every=50,
        step_size=1,
        show_progress=True,
    ):
        """
        Executes thetime evolution for `n_iter` steps of `dt` and writing values of
        the observables `obs` to the output. The output are JSON file at
        `output_prefix.{log,state}`, overwriting files with the same prefix.

        Args:
            :output_prefix: The prefix at which JSON output should be stored.
            :n_iter: the total number of iterations
            :obs: An iterable containing all observables that should be computed
            :save_params_every: Every how many steps the parameters of the network should be
            serialized to disk (ignored if logger is provided)
            :write_every: Every how many steps the json data should be flushed to disk (ignored if
            logger is provided)
            :step_size: Every how many steps should observables be logged to disk (default=1)
            :show_progress: If true displays a progress bar (default=True)
        """
        if obs is None:
            obs = {}

        logger = _JsonLog(output_prefix, save_params_every, write_every)

        # Don't log on non-root nodes
        if self._mpi_nodes != 0:
            logger = None

        with tqdm(
            self.iter(n_iter, step_size), total=n_iter, disable=not show_progress
        ) as itr:
            for step in itr:
                # if the cost-function is defined then report it in the progress bar
                energy = self.estimate(self._h(self.t))
                itr.set_postfix_str(
                    "t={:.2e}, Energy={:.6e}".format(self.t, energy.mean.real)
                )

                obs_data = self.estimate(obs)
                obs_data["Energy"] = energy

                log_data = {}
                if self._loss_stats is not None:
                    obs_data[self._loss_name] = self._loss_stats
                log_data["Time"] = self.t

                if logger is not None:
                    logger(step, log_data, _MockMachine(self.state))

        # flush at the end of the evolution so that final values are saved to
        # file
        if logger is not None:
            logger.flush(_MockMachine(self.state))



from .. import AbstractMachine
import numpy as _np


class Diagonal(AbstractMachine):
    def __init__(self, density_matrix):
        self._rho = density_matrix
        super().__init__(
            density_matrix.hilbert,
            dtype=density_matrix.dtype,
        )

    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The length of `out` should be `x.shape[0]`.

        Returns:
            A complex number when `x` is a vector and vector when `x` is a
            matrix.
        """
        return self._rho.log_val(x, x, out)

    def der_log(self, x, out=None):
        return self._rho.der_log(x, x, out)

    @property
    def n_par(self):
        r"""The number of variational parameters in the machine."""
        return self._rho.n_par

    @property
    def state_dict(self):
        r"""A dictionary containing the parameters of this machine"""
        return self._rho.state_dict

    @property
    def parameters(self):
        r"""The parameters of the machine."""
        return self._rho.parameters



from .diagonal import Diagonal
from .abstract_density_matrix import AbstractDensityMatrix
from .rbm import RbmSpin

from netket.utils import jax_available


from .jax import Jax, NdmSpin, NdmSpinPhase, JaxRbmSpin
from .flax import Flax



from netket.legacy.machine import Jax as JaxMachine
from ..metropolis_hastings import MetropolisHastings, MetropolisHastingsPt
from .metropolis_hastings import MetropolisHastings as JaxMetropolisHastings


@MetropolisHastings.register(JaxMachine)
def _JaxMetropolisHastings(machine, kernel, n_chains=16, sweep_size=None, rng_key=None):
    return JaxMetropolisHastings(machine, kernel, n_chains, sweep_size, rng_key)


@MetropolisHastingsPt.register(JaxMachine)
def _JaxMetropolisHastingsPt(
    machine, kernel, n_replicas=32, sweep_size=None, rng_key=None
):
    raise NotImplementedError("Parallel tempering samplers not yet implemented in Jax")


# Register Jax kernels here
from .._kernels import _LocalKernel, _ExchangeKernel, _HamiltonianKernel, _CustomKernel
from .local_kernel import _JaxLocalKernel
from .exchange_kernel import _JaxExchangeKernel


@_LocalKernel.register(JaxMachine)
def _Jax_LocalKernel(machine):
    return _JaxLocalKernel(machine.hilbert.local_states, machine.input_size)


@_ExchangeKernel.register(JaxMachine)
def _Jax_ExchangeKernel(machine, clusters):
    return _JaxExchangeKernel(machine.hilbert, clusters)


@_HamiltonianKernel.register(JaxMachine)
def _Jax_HamiltonianKernel(machine, hamiltonian):
    raise NotImplementedError("Kernel not yet implemented in Jax")


@_CustomKernel.register(JaxMachine)
def _Jax_CustomKernel(machine, move_operators, move_weights=None):
    raise NotImplementedError("Kernel not yet implemented in Jax")



import numpy as _np
from numba import jit, objmode
from netket.utils.mpi import (
    n_nodes as _n_nodes,
    node_number as _rank,
)


@jit
def seed(seed=None):
    """Seed the random number generator. Each MPI process is automatically assigned
    a different, process-dependent, sub-seed.

    Parameters:
              seed (int, optional): Seed for the randon number generator.

    """

    if seed == None:
        seed = _np.random.randint(0, 1 << 32)

    _np.random.seed(seed)
    derived_seed = _np.random.randint(0, 1 << 32, size=_n_nodes)

    _np.random.seed(derived_seed[_rank])


@jit
def uniform(low=0.0, high=1.0):
    """
    Draw samples from a uniform distribution. Samples are uniformly distributed
    over the half-open interval [low, high) (includes low, but excludes high).

    Parameters:
              low (float, optional): Lower boundary of the output interval.
                                     All values generated will be greater than
                                     or equal to low. The default value is 0.
              high (float, optional): Upper boundary of the output interval.
                                     All values generated will be less than high.
                                     The default value is 1.0.
    Returns:
              float: A randon number uniformly distributed in [low,high).

    """
    return _np.random.uniform(low, high)


@jit
def randint(low, high, size=()):
    """
    Generate random integers from low (inclusive) to high (exclusive).

    Args:
        low (int): Lowest (signed) integer to be drawn from the distribution.
        high (int): One above the largest (signed) integer to be drawn from the distribution.
        size: Output shape.

    Returns:
        int: A random integer uniformely distributed in [low,high).

    """
    # Ugly workaround for numba not being able to jit compile randint with size arg.
    # size is also a required argument for numba reasons.
    x = _np.random.rand(*size)
    return _np.asarray(_np.floor(x * (high - low)) + low, dtype=_np.int64)


def choice(a, size=None, replace=True, p=None):
    # TODO use always numpy version when argument p is made available in numba
    """
    Generates a random sample from a given 1-D array.

    Args:
        a (1-D array-like): A random sample is generated from its elements.
        replace (boolean, optional): Whether the sample is with or without replacement.
        p (1-D array-like, optional): The probabilities associated with each entry in a.
                    If not given the sample assumes a uniform distribution over all entries in a.

    Returns:
        single item or ndarry: The generated random samples
    """
    return _np.random.choice(a, size, replace, p)


# By default, the generator is initialized with a random seed (on node 0)
# and then propagated correctly to the other nodes
seed(None)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
import numpy as np
import jax

# 1D Lattice
L = 20

g = nk.graph.Chain(length=L, pbc=True)
# Lattice translation operations

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes, total_sz=0)

# Ising spin hamiltonian
ha = nk.operator.Heisenberg(hilbert=hi, graph=g)

# RBM Spin Machine
ma = nk.models.RBMSymm(
    symmetries=g.translations(),
    alpha=4,
    use_visible_bias=False,
    use_hidden_bias=True,
    dtype=float,
)

# Metropolis Local Sampling
sa = nk.sampler.MetropolisExchange(hi, graph=g, n_chains=16)

# Optimizer
op = nk.optim.Sgd(learning_rate=0.01)
sr = nk.optim.SR(diag_shift=0.1)

# Variational monte carlo driver
gs = nk.VMC(ha, op, sa, ma, n_samples=1000, n_discard_per_chain=100, preconditioner=sr)

# Print parameter structure
print(f"# variational parameters: {gs.state.n_parameters}")


# Run the optimization for 300 iterations
gs.run(n_iter=300, out="test")



from functools import singledispatch
from . import numpy


@singledispatch
def AmsGrad(machine, learning_rate=0.001, beta1=0.9, beta2=0.999, epscut=1.0e-7):
    r"""AmsGrad Optimizer.
        In some cases, adaptive learning rate methods such as AdaMax fail
        to converge to the optimal solution because of the exponential
        moving average over past gradients. To address this problem,
        Sashank J. Reddi, Satyen Kale and Sanjiv Kumar proposed the
        AmsGrad [update algorithm](https://openreview.net/forum?id=ryQu7f-RZ).
        The update rule for :math:`\mathbf{v}` (equivalent to :math:`E[g^2]` in AdaDelta
        and :math:`\mathbf{s}` in RMSProp) is modified such that :math:`v^\prime_k \geq v_k`
        is guaranteed, giving the algorithm a "long-term memory" of past gradients.
        The vectors :math:`\mathbf{m}` and :math:`\mathbf{v}` are initialized to zero, and
        are updated with the parameters :math:`\mathbf{p}`:

        .. math:: m^\prime_k &= \beta_1 m_k + (1-\beta_1)G_k(\mathbf{p})\\
                  v^\prime_k &= \beta_2 v_k + (1-\beta_2)G_k(\mathbf{p})^2\\
                  v^\prime_k &= \mathrm{Max}(v^\prime_k, v_k)\\
                  p^\prime_k &= p_k - \frac{\eta}{\sqrt{v^\prime_k}+\epsilon}m^\prime_k


        Constructs a new ``AmsGrad`` optimizer.

        Args:
           learning_rate: The learning rate $\eta$.
           beta1: First exponential decay rate.
           beta2: Second exponential decay rate.
           epscut: Small epsilon cutoff.

        Examples:
           Simple AmsGrad optimizer.

           >>> from netket.optimizer import AmsGrad
           >>> op = AmsGrad()
    """

    return numpy.AmsGrad(learning_rate=0.001, beta1=0.9, beta2=0.999, epscut=1.0e-7)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
import optax
import pytest

from .. import common

pytestmark = common.skipif_mpi


@pytest.mark.parametrize(
    "hilbert",
    [
        pytest.param(
            nk.hilbert.Spin(s=1 / 2, N=4),
            id="spin_1/2",
        ),
        pytest.param(
            nk.hilbert.Spin(s=1, N=4),
            id="spin_1",
        ),
        pytest.param(
            nk.hilbert.Fock(n_max=3, N=4),
            id="fock",
        ),
    ],
)
def test_AR_VMC(hilbert):
    graph = nk.graph.Hypercube(length=hilbert.size, n_dim=1)
    model = nk.models.ARNNDense(hilbert=hilbert, layers=3, features=5)
    sampler = nk.sampler.ARDirectSampler(hilbert, n_chains=3)

    vstate = nk.vqs.MCState(sampler, model, n_samples=6)
    assert vstate.n_discard_per_chain == 0
    vstate.sample()

    H = nk.operator.Ising(hilbert=hilbert, graph=graph, h=1)
    optimizer = optax.adam(learning_rate=1e-3)
    vmc = nk.VMC(H, optimizer, variational_state=vstate)
    vmc.run(n_iter=3)



import pytest

import netket as nk

from .. import common

pytestmark = common.skipif_mpi


def test_deprecated_sr():
    with pytest.warns(FutureWarning):
        nk.optimizer.sr.SRLazyCG()

    with pytest.warns(FutureWarning):
        nk.optimizer.sr.SRLazyGMRES()



from ..abstract_optimizer import AbstractOptimizer
import numpy as _np


class AdaDelta(AbstractOptimizer):
    r"""AdaDelta Optimizer.
        Like RMSProp, `AdaDelta <http://arxiv.org/abs/1212.5701>`_ corrects the
        monotonic decay of learning rates associated with AdaGrad,
        while additionally eliminating the need to choose a global
        learning rate :math:`\eta`. The NetKet naming convention of
        the parameters strictly follows the one introduced in the original paper;
        here :math:`E[g^2]` is equivalent to the vector :math:`\mathbf{s}` from RMSProp.
        :math:`E[g^2]` and :math:`E[\Delta x^2]` are initialized as zero vectors.

        .. math:: E[g^2]^\prime_k &= \rho E[g^2] + (1-\rho)G_k(\mathbf{p})^2\\
                  \Delta p_k &= - \frac{\sqrt{E[\Delta x^2]+\epsilon}}{\sqrt{E[g^2]+ \epsilon}}G_k(\mathbf{p})\\
                  E[\Delta x^2]^\prime_k &= \rho E[\Delta x^2] + (1-\rho)\Delta p_k^2\\
                  p^\prime_k &= p_k + \Delta p_k
    """

    def __init__(self, rho=0.95, epscut=1.0e-7, l2reg=0):
        r"""
        Constructs a new ``AdaDelta`` optimizer.

        Args:
           rho: Exponential decay rate, in [0,1].
           epscut: Small :math:`\epsilon` cutoff.
           l2reg (float): The amount of L2 regularization.

        Examples:
           Simple AdaDelta optimizer.

           >>> from netket.optimizer import AdaDelta
           >>> op = AdaDelta()
        """

        self._rho = rho
        self._epscut = epscut
        self._l2reg = l2reg

        if epscut <= 0:
            raise ValueError("Invalid epsilon cutoff.")
        if l2reg < 0:
            raise ValueError("Invalid L2 regularization.")
        if rho < 0 or rho > 1:
            raise ValueError("Invalid beta.")

        self._Eg2 = None
        self._Edx2 = None

    def update(self, grad, pars):
        if self._Eg2 is None:
            self._Eg2 = _np.zeros(pars.shape[0])
            self._Edx2 = _np.zeros(pars.shape[0])

        self._Eg2 = self._rho * self._Eg2 + (1.0 - self._rho) * _np.absolute(grad) ** 2

        Dx = _np.sqrt(self._Edx2 + self._epscut) * grad
        Dx /= _np.sqrt(self._Eg2 + self._epscut)

        pars -= Dx

        self._Edx2 = self._rho * self._Edx2 + (1.0 - self._rho) * _np.absolute(Dx) ** 2

        return pars

    def reset(self):
        if self._mt is not None:
            self._Eg2.fill(0.0)
            self._Edx2.fill(0.0)

    def __repr__(self):
        rep = "AdaDelta optimizer with these parameters :"
        rep += "\nRho = " + str(self._rho)
        rep += "\nepscut = " + str(self._epscut)
        return rep



from functools import singledispatch
import numpy as _np

from netket.utils.mpi import mpi_available as _mpi_available, n_nodes as _n_nodes

if _mpi_available:
    from netket.utils.mpi import MPI_py_comm as _MPI_py_comm
    from netket.utils.mpi import MPI_jax_comm as _MPI_jax_comm
    from netket.utils.mpi import MPI as _MPI


@singledispatch
def sum_inplace(x):
    """
    Computes the elementwise sum of an array or a scalar across all MPI processes.
    Attempts to perform this sum inplace if possible, but for some types a copy
    might be returned.

    Args:
        a: The input array, which will usually be overwritten in place.
    Returns:
        out: The reduced array.
    """
    raise TypeError("Unknown type to perform dispatch upon: {}".format(type(x)))


#######
# Scalar
@sum_inplace.register(complex)
@sum_inplace.register(_np.float64)
@sum_inplace.register(_np.float32)
@sum_inplace.register(_np.complex64)
@sum_inplace.register(_np.complex128)
@sum_inplace.register(float)
@sum_inplace.register(int)
def sum_inplace_scalar(a):
    ar = _np.asarray(a)

    if _n_nodes > 1:
        _MPI_py_comm.Allreduce(_MPI.IN_PLACE, ar.reshape(-1), op=_MPI.SUM)

    return ar


##############
# Numpy Array
#
@sum_inplace.register(_np.ndarray)
def sum_inplace_MPI(a):
    """
    Computes the elementwise sum of a numpy array over all MPI processes.

    Args:
        a (numpy.ndarray): The input array, which will be overwritten in place.
    """
    if _n_nodes > 1:
        _MPI_py_comm.Allreduce(_MPI.IN_PLACE, a.reshape(-1), op=_MPI.SUM)

    return a


##############
# Jax
#
from netket.utils import jax_available, mpi4jax_available

if jax_available:
    import numpy as _np
    import jax

    if mpi4jax_available:
        import mpi4jax

        @sum_inplace.register(jax.interpreters.xla.DeviceArray)
        @sum_inplace.register(jax.interpreters.partial_eval.JaxprTracer)
        @sum_inplace.register(jax.interpreters.partial_eval.DynamicJaxprTracer)
        @sum_inplace.register(jax.interpreters.ad.JVPTracer)
        def sum_inplace_jax(x):
            if _n_nodes == 1:
                return x
            else:
                # Note: We must supply a token because we can't transpose `create_token`.
                # The token can't depend on x for the same reason
                # This token depends on a constant and will be eliminated by DCE
                token = jax.lax.create_token(0)
                res, _ = mpi4jax.allreduce(
                    x, op=_MPI.SUM, comm=_MPI_jax_comm, token=token
                )
                return res

    else:

        @sum_inplace.register(jax.interpreters.xla.DeviceArray)
        @sum_inplace.register(jax.interpreters.partial_eval.JaxprTracer)
        @sum_inplace.register(jax.interpreters.partial_eval.DynamicJaxprTracer)
        @sum_inplace.register(jax.interpreters.ad.JVPTracer)
        def sum_inplace_jax(x):
            if _n_nodes == 1:
                return x
            else:
                raise RuntimeError(
                    "The package mpi4jax is required in order to use jax machines with SR and\
                    more than one MPI process. To solve this issue, run `pip install mpi4jax` and restart netket."
                )



import warnings
from jax.tree_util import tree_map


# Higher-level VMC functions:


def estimate_expectations(
    ops, sampler, n_samples, n_discard=None, compute_gradients=False
):
    """
    For a sequence of linear operators, computes a statistical estimate of the
    respective expectation values, variances, and optionally gradients of the
    expectation values with respect to the variational parameters.

    The estimate is based on `n_samples` configurations
    obtained from `sampler`.

    Args:
        ops: pytree of linear operators
        sampler: A NetKet sampler
        n_samples: Number of MC samples used to estimate expectation values
        n_discard: Number of MC samples dropped from the start of the
            chain (burn-in). Defaults to `n_samples //10`.
        compute_gradients: Whether to compute the gradients of the
            observables.

    Returns:
        Either `stats` or, if `der_logs` is passed, a tuple of `stats` and `grad`:
            stats: A sequence of Stats object containing mean, variance,
                and MC diagonstics for each operator in `ops`.
            grad: A sequence of gradients of the expectation value of `op`,
                  as ndarray of shape `(psi.n_par,)`, for each `op` in `ops`.
    """

    from netket.operator import local_values as _local_values
    from netket.legacy.stats import (
        statistics as _statistics,
        mean as _mean,
        sum_inplace as _sum_inplace,
    )

    psi = sampler.machine

    if not n_discard:
        n_discard = n_samples // 10

    # Burnout phase
    sampler.generate_samples(n_discard)
    # Generate samples
    samples = sampler.generate_samples(n_samples).reshape(
        (-1, sampler.sample_shape[-1])
    )

    def estimate(op):
        lvs = _local_values(op, psi, samples)
        stats = _statistics(lvs.T)

        if compute_gradients:
            samples_r = samples.reshape((-1, samples.shape[-1]))
            eloc_r = (lvs - _mean(lvs)).reshape(-1, 1)
            grad = sampler.machine.vector_jacobian_prod(
                samples_r,
                eloc_r / n_samples,
            )
            return stats, grad
        else:
            return stats

    return tree_map(estimate, ops)



from ..abstract_optimizer import AbstractOptimizer
import numpy as _np


class Momentum(AbstractOptimizer):
    r"""Momentum-based Optimizer.
        The momentum update incorporates an exponentially weighted moving average
        over previous gradients to speed up descent
        `Qian, N. (1999) <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.57.5612&rep=rep1&type=pdf>`_.
        The momentum vector :math:`\mathbf{m}` is initialized to zero.
        Given a stochastic estimate of the gradient of the cost function
        :math:`G(\mathbf{p})`, the updates for the parameter :math:`p_k` and
        corresponding component of the momentum :math:`m_k` are

        .. math:: m^\prime_k &= \beta m_k + (1-\beta)G_k(\mathbf{p})\\
        p^\prime_k &= \eta m^\prime_k
    """

    def __init__(self, learning_rate, beta=0.9, l2reg=0):
        r"""
        Constructs a new ``Momentum`` optimizer.

        Args:
            learning_rate (float): The learning rate :math:`\eta`
            beta (float): Momentum exponential decay rate, should be in [0,1].
            l2reg (float): The amount of L2 regularization.

        Examples:
            Momentum optimizer.

            >>> from netket.optimizer import Momentum
            >>> op = Momentum(learning_rate=0.01)
        """
        self._learning_rate = learning_rate
        self._l2reg = l2reg
        self._beta = beta
        self._eta = learning_rate
        self._mt = None

        if learning_rate <= 0:
            raise ValueError("Invalid learning rate.")
        if l2reg < 0:
            raise ValueError("Invalid L2 regularization.")
        if beta < 0 or beta > 1:
            raise ValueError("Invalid beta.")

    def update(self, grad, pars):
        if self._mt is None:
            self._mt = _np.zeros(pars.shape[0], dtype=pars.dtype)

        self._mt = self._beta * self._mt + (1.0 - self._beta) * grad
        pars -= self._eta * (self._mt + self._l2reg * pars)

        return pars

    def reset(self):
        if self._mt is not None:
            self._mt.fill(0.0)

    def __repr__(self):
        rep = "Momentum optimizer with these parameters :"
        rep += "\nLearning Rate = " + str(self._learning_rate)
        rep += "\nL2 Regularization = " + str(self._l2reg)
        rep += "\nBeta = " + str(self._beta)
        return rep



from . import AbstractDensityMatrix
from .. import RbmSpin as PureRbmSpin
import numpy as _np


class RbmSpin(AbstractDensityMatrix):
    def __init__(
        self,
        hilbert,
        n_hidden=None,
        alpha=None,
        use_visible_bias=True,
        use_hidden_bias=True,
        automorphisms=None,
        dtype=complex,
    ):
        super().__init__(hilbert, dtype=dtype)

        if automorphisms is not None:
            if isinstance(automorphisms, netket.graph.AbstractGraph):
                automorphisms = automorphisms.automorphisms()
            import itertools

            automorphisms = [
                prod[0] + prod[1] for prod in itertools.product(autom, autom)
            ]

        input_like = _np.zeros(hilbert.size * 2)
        self._prbm = PureRbmSpin(
            input_like,
            n_hidden,
            alpha,
            use_visible_bias,
            use_hidden_bias,
            automorphisms,
            dtype,
        )
        self._plog_val = self._prbm.log_val
        self._pder_log = self._prbm.der_log

    def log_val(self, xr, xc=None, out=None):
        r"""Computes the logarithm of the density matrix for a batch of visible
        quantum numbers `(xr,xc)` and stores the result into `out`.
        Specifically, for each element of the batch i, this function should compute
        out[i]=log(rho(xr[i],xc[i])).
        If xr is None, it is assumed that xr has twice as many quantum numbers and
        contains both row and columns, stored contigously.

        Args:
            xr: A matrix of `float64` of shape `(*, self.n_visible)` if xc is given.
                If xc is None, then this should be a matrix of `float64` of shape `(*, 2*self.n_visible)`.
            xc (optional): A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The length of `out` should be `xr.shape[0]`.

        Returns:
            A vector out[i]=log(rho(xr[i],xc[i])).
        """
        if xc is None:
            return self._plog_val(xr, out)
        else:
            return self._plog_val(_np.hstack((xr, xc)), out)

    def der_log(self, xr, xc=None, out=None):
        r"""Computes the gradient of the logarithm of the density matrix for a
        batch of visible configurations `(xr,xc)` and stores the result into `out`.

        Args:
            xr: A matrix of `float64` of shape `(*, self.n_visible)` if xc is given.
                If xc is None, then this should be a matrix of `float64` of shape `(*, 2*self.n_visible)`.
            xc (optional): A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`.
                `out` should be a matrix of shape `(x.shape[0], self.n_par)`.

        Returns:
            `out`
        """
        if xc is None:
            return self._pder_log(xr, out)
        else:
            return self._pder_log(_np.hstack((xr, xc)), out)

    @property
    def state_dict(self):
        r"""A dictionary containing the parameters of this machine"""
        return self._prbm.state_dict



import numpy as np
from pytest import approx, raises
from scipy.integrate import solve_ivp
from scipy.linalg import norm

import netket.legacy as nk
from netket.exact import PyExactTimePropagation

import pytest

pytestmark = pytest.mark.legacy

ATOL = 1e-9
RTOL = 1e-9
TIME = 20.0


def _setup_model():
    L = 8
    hi = nk.hilbert.Spin(0.5) ** L
    ham = nk.operator.Heisenberg(hi, nk.graph.Hypercube(L, 1))

    psi0 = np.random.rand(hi.n_states) + 1j * np.random.rand(hi.n_states)
    psi0 /= norm(psi0)

    return ham, psi0


def overlap(phi, psi):
    return np.abs(np.vdot(phi, psi)) ** 2 / (np.vdot(phi, phi) * np.vdot(psi, psi)).real


def test_python_real_time_evolution():
    ham, psi0 = _setup_model()
    ham_matrix = ham.to_linear_operator()

    driver = PyExactTimePropagation(
        ham,
        dt=TIME,
        t0=0.0,
        initial_state=psi0,
        propagation_type="real",
        solver_kwargs={"atol": ATOL, "rtol": RTOL},
    )

    assert np.allclose(driver.state, psi0)

    driver.advance(1)
    assert driver.t == TIME

    def rhs_real(t, x):
        return -1.0j * ham_matrix.dot(x)

    res = solve_ivp(rhs_real, (0.0, TIME), psi0, atol=ATOL, rtol=RTOL)
    psi_scipy = res.y[:, -1]

    assert driver.t == approx(res.t[-1])
    assert overlap(driver.state, psi_scipy) == approx(1.0)


def test_python_imag_time_evolution():
    ham, psi0 = _setup_model()
    ham_matrix = ham.to_linear_operator()

    driver = PyExactTimePropagation(
        ham,
        dt=TIME,
        t0=0.0,
        initial_state=psi0,
        propagation_type="imaginary",
        solver_kwargs={"atol": ATOL, "rtol": RTOL},
    )

    assert np.allclose(driver.state, psi0)

    driver.advance(1)
    assert driver.t == TIME

    def rhs_imag(t, x):
        mean = np.vdot(x, ham_matrix.dot(x))
        return -ham_matrix.dot(x) + mean * x

    res = solve_ivp(rhs_imag, (0.0, TIME), psi0, atol=ATOL, rtol=RTOL)
    psi_scipy = res.y[:, -1]

    assert driver.t == approx(res.t[-1])
    assert overlap(driver.state, psi_scipy) == approx(1.0)


def test_time_stepping():
    ham, psi0 = _setup_model()

    driver = PyExactTimePropagation(
        ham,
        t0=0.0,
        dt=0.05,
        initial_state=psi0,
        propagation_type="real",
        solver_kwargs={"atol": 1e-2, "rtol": 1e-2},
    )
    ts = [driver.t for _ in driver.iter(21)]

    ts_ref = np.linspace(0, 1, 21, endpoint=True)
    assert np.allclose(ts, ts_ref)



import numpy as np
import matplotlib.pyplot as plt
import json

plt.ion()

# J2/J1=0.2
# exact = -32.8116650087338

# J2/J1=0.4
# exact=-30.5017311953

# J2/J1=0.5
exact = -30

# J2/J1=0.6
# exact = -30.47051755167271

# J2/J1=0.8
# exact = -33.84508963604603

while True:
    plt.clf()
    plt.ylabel("Energy")
    plt.xlabel("Iteration #")

    iters = []
    energy = []
    sigma = []
    evar = []

    data = json.load(open("test.log"))
    for iteration in data["Output"]:
        iters.append(iteration["Iteration"])
        energy.append(iteration["Energy"]["Mean"])
        sigma.append(iteration["Energy"]["Sigma"])
        evar.append(iteration["Energy"]["Variance"])

    plt.errorbar(iters, energy, yerr=sigma, color="red")
    plt.axhline(y=exact, xmin=0, xmax=iters[-1], linewidth=2, color="k", label="Exact")
    nres = len(iters)
    if nres > 100:
        fitx = iters[-100:-1]
        fity = energy[-100:-1]
        z = np.polyfit(fitx, fity, deg=0)
        p = np.poly1d(z)

        plt.plot(fitx, p(fitx))
        if nres > 100:
            plt.xlim([nres - 100, nres])
            maxval = np.max(energy[-100:-1])
            plt.ylim([exact - (np.abs(exact) * 0.01), maxval + np.abs(maxval) * 0.01])
        error = (z[0] - exact) / -exact
        plt.gca().text(
            0.95,
            0.8,
            "Relative Error : " + "{:.2e}".format(error),
            verticalalignment="bottom",
            horizontalalignment="right",
            color="green",
            fontsize=15,
            transform=plt.gca().transAxes,
        )

    plt.legend(frameon=False)
    plt.pause(1)


plt.show()



from functools import singledispatch
from . import numpy


@singledispatch
def RmsProp(machine, learning_rate=0.001, beta=0.9, epscut=1.0e-7):
    r"""RMSProp optimizer.

    RMSProp is a well-known update algorithm proposed by Geoff Hinton
    in his Neural Networks course notes `Neural Networks course notes
    <http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.
    It corrects the problem with AdaGrad by using an exponentially weighted
    moving average over past squared gradients instead of a cumulative sum.
    After initializing the vector :math:`\mathbf{s}` to zero, :math:`s_k` and t
    he parameters :math:`p_k` are updated as

    .. math:: s^\prime_k = \beta s_k + (1-\beta) G_k(\mathbf{p})^2 \\
              p^\prime_k = p_k - \frac{\eta}{\sqrt{s_k}+\epsilon} G_k(\mathbf{p})


    Constructs a new ``RmsProp`` optimizer.

    Args:
       learning_rate: The learning rate :math:`\eta`
       beta: Exponential decay rate.
       epscut: Small cutoff value.

    Examples:
       RmsProp optimizer.

       >>> from netket.optimizer import RmsProp
       >>> op = RmsProp(learning_rate=0.02)
    """
    return numpy.RmsProp(learning_rate, beta, epscut)



from .metropolis_hastings import *
from ._kernels import _LocalKernel


def MetropolisLocal(machine, n_chains=16, sweep_size=None, **kwargs):
    r"""
    Sampler acting on one local degree of freedom.

    This sampler acts locally only on one local degree of freedom :math:`s_i`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s_N`,
    where :math:`s^\prime_i \\neq s_i`.

    The transition probability associated to this
    sampler can be decomposed into two steps:

    1. One of the site indices :math:`i = 1\dots N` is chosen
    with uniform probability.
    2. Among all the possible (:math:`m`) values that :math:`s_i` can take,
    one of them is chosen with uniform probability.

    For example, in the case of spin :math:`1/2` particles, :math:`m=2`
    and the possible local values are :math:`s_i = -1,+1`.
    In this case then :class:`MetropolisLocal` is equivalent to flipping a random spin.

    In the case of bosons, with occupation numbers
    :math:`s_i = 0, 1, \dots n_{\mathrm{max}}`, :class:`MetropolisLocal`
    would pick a random local occupation number uniformly between :math:`0`
    and :math:`n_{\mathrm{max}}`.

    Args:
        machine: A machine :math:`\Psi(s)` used for the sampling.
                 The probability distribution being sampled
                 from is :math:`F(\Psi(s))`, where the function
                 :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.

        n_chains:   The number of Markov Chain to be run in parallel on a single process.
        sweep_size: The number of exchanges that compose a single sweep.
                    If None, sweep_size is equal to the number of degrees of freedom (n_visible).


    Examples:
        Sampling from a RBM machine in a 1D lattice of spin 1/2

        >>> import netket as nk
        >>>
        >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
        >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
        >>>
        >>> # RBM Spin Machine
        >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
        >>>
        >>> # Construct a MetropolisLocal Sampler
        >>> sa = nk.sampler.MetropolisLocal(machine=ma)
        >>> print(sa.machine.hilbert.size)
        100
    """

    return MetropolisHastings(
        machine, _LocalKernel(machine), n_chains, sweep_size, **kwargs
    )


def MetropolisLocalPt(machine, n_replicas=16, sweep_size=None, **kwargs):
    r"""
    This sampler performs parallel-tempering
    moves in addition to the local moves implemented in `MetropolisLocal`.
    The number of replicas can be chosen by the user.

    Args:
         machine: A machine :math:`\Psi(s)` used for the sampling.
                  The probability distribution being sampled
                  from is :math:`F(\Psi(s))`, where the function
                  :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
         n_replicas: The number of replicas used for parallel tempering.
         sweep_size: The number of exchanges that compose a single sweep.
                     If None, sweep_size is equal to the number of degrees of freedom (n_visible).

    """
    return MetropolisHastingsPt(
        machine, _LocalKernel(machine), n_replicas, sweep_size, **kwargs
    )



from netket.utils import jax_available


def info(obj, depth=None):
    if hasattr(obj, "info"):
        return obj.info(depth)
    else:
        return str(obj)


if jax_available:
    from jax import tree_map as _tree_map

    def tree_map(fun, tree, *args, **kwargs):
        return _tree_map(lambda x: fun(x, *args, **kwargs), tree)


else:

    def tree_map(fun, tree, *args, **kwargs):
        """
        Maps all the leafs in the tree, applying the function with the leave as first
        positional argument.
        Any additional argument after the first two is forwarded to the function call.

        Args:
            fun: the function to apply to all leafs
            tree: the structure containing leafs. This can also be just a leaf
            *args: additional positional arguments passed to fun
            **kwargs: additional kw arguments passed to fun

        Returns:
            An equivalent tree, containing the result of the function call.
        """
        if tree is None:
            return None
        elif isinstance(tree, list):
            return [tree_map(fun, val, *args, **kwargs) for val in tree]
        elif isinstance(tree, tuple):
            if not hasattr(tree, "_fields"):
                # If it is not a namedtuple, recreate it as a tuple.
                return tuple(tree_map(fun, val, *args, **kwargs) for val in tree)
            else:
                # If it is a namedtuple, than keep that type information.
                return type(tree)(
                    *(tree_map(fun, val, *args, **kwargs) for val in tree)
                )
        elif isinstance(tree, dict):
            return {
                key: tree_map(fun, value, *args, **kwargs)
                for key, value in tree.items()
            }
        else:
            return fun(tree, *args, **kwargs)


if jax_available:
    from jax import tree_multimap

    def trees2_map(fun, tree1, tree2, *args, **kwargs):
        return tree_multimap(lambda x, y: fun(x, y, *args, **kwargs), tree1, tree2)


else:

    def trees2_map(fun, tree1, tree2, *args, **kwargs):
        """
        Maps all the leafs in the two trees, applying the function with the leafs of tree1
        as first argument and the leafs of tree2 as second argument
        Any additional argument after the first two is forwarded to the function call.

        This is usefull e.g. to sum the leafs of two trees

        Args:
            fun: the function to apply to all leafs
            tree1: the structure containing leafs. This can also be just a leaf
            tree2: the structure containing leafs. This can also be just a leaf
            *args: additional positional arguments passed to fun
            **kwargs: additional kw arguments passed to fun

        Returns:
            An equivalent tree, containing the result of the function call.
        """
        if tree1 is None:
            return None
        elif isinstance(tree1, list):
            return [
                trees2_map(fun, val1, val2, *args, **kwargs)
                for val1, val2 in zip(tree1, tree2)
            ]
        elif isinstance(tree1, tuple):
            return tuple(
                trees2_map(fun, val1, val2, *args, **kwargs)
                for val1, val2 in zip(tree1, tree2)
            )
        elif isinstance(tree1, dict):
            return {
                key: trees2_map(fun, val1, val2, *args, **kwargs)
                for (key, val1), (key2, val2) in zip(tree1.items(), tree2.items())
            }
        else:
            return fun(tree1, tree2, *args, **kwargs)


from netket.utils import jax_available

if jax_available:
    import jax
    from jax.tree_util import tree_flatten, tree_unflatten
    import jax.numpy as jnp

    @jax.jit
    def jax_shape_for_update(update, shape_like):
        r"""Reshapes grads from array to tree like structure if neccesary for update

        Args:
            grads: a 1d jax/numpy array
            shape_like: this as in instance having the same type and shape of
                        the desired conversion.

        Returns:
            A possibly non-flat structure of jax arrays containing a copy of data
            compatible with the given shape if jax_available and a copy of update otherwise
        """

        shf, tree = tree_flatten(shape_like)

        updatelist = []
        k = 0
        for s in shf:
            size = s.size
            updatelist.append(jnp.asarray(update[k : k + size]).reshape(s.shape))
            k += size

        return tree_unflatten(tree, updatelist)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import random

from .abstract_density_matrix import AbstractDensityMatrix
from ..flax import Flax as FlaxPure
from functools import partial


class Flax(FlaxPure, AbstractDensityMatrix):
    def __init__(self, hilbert, module, dtype=complex):
        """
        Wraps a stax network (which is a tuple of `init_fn` and `predict_fn`)
        so that it can be used as a NetKet density matrix.

        Args:
            hilbert: Hilbert space on which the state is defined. Should be a
                subclass of `netket.hilbert.Hilbert`.
            module: A pair `(init_fn, predict_fn)`. See the documentation of
                `jax.experimental.stax` for more info.
            dtype: either complex or float, is the type used for the weights.
                In both cases the module must have a single output.
        """
        AbstractDensityMatrix.__init__(self, hilbert, dtype)
        FlaxPure.__init__(self, hilbert, module)

        assert self.input_size == self.hilbert.size * 2

    @staticmethod
    @jax.jit
    def _dminput(xr, xc):
        if xc is None:
            x = xr
        else:
            x = jnp.hstack((xr, xc))
        return x

    def log_val(self, xr, xc=None, out=None):
        x = self._dminput(xr, xc)

        return FlaxPure.log_val(self, x, out=out)

    def der_log(self, xr, xc=None, out=None):
        x = self._dminput(xr, xc)

        return FlaxPure.der_log(self, x, out=out)

    def diagonal(self):
        from .diagonal import Diagonal

        diag = Diagonal(self)

        def diag_jax_forward(params, x):
            return self.jax_forward(params, self._dminput(x, x))

        diag.jax_forward = diag_jax_forward

        return diag



from .ada_delta import AdaDelta
from .ada_grad import AdaGrad
from .ada_max import AdaMax
from .ams_grad import AmsGrad
from .momentum import Momentum
from .rms_prop import RmsProp
from .sgd import Sgd

from .stochastic_reconfiguration import SR



from .abstract_machine import AbstractMachine

from .rbm import RbmSpin, RbmSpinReal, RbmSpinSymm, RbmMultiVal, RbmSpinPhase
from .jastrow import Jastrow, JastrowSymm

from netket.utils import torch_available

from .jax import Jax, JaxRbm, MPSPeriodic, JaxRbmSpinPhase
from .jax import DenseReal, SumLayer, LogCoshLayer
from .flax import Flax

if torch_available:
    from .torch import Torch, TorchLogCosh, TorchView


from . import density_matrix



import json
from pytest import approx
import netket.legacy as nk
import numpy as np
import shutil
import tempfile

import pytest

pytestmark = pytest.mark.legacy


SEED = 3141592


def _setup_vmc(
    n_samples=200, diag_shift=0, use_iterative=False, lsq_solver=None, **kwargs
):
    L = 8
    nk.random.seed(SEED)
    hi = nk.hilbert.Spin(s=0.5) ** L

    ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
    ma.init_random_parameters(sigma=0.01, seed=SEED)

    ha = nk.operator.Ising(hi, nk.graph.Hypercube(length=L, n_dim=1), h=1.0)
    sa = nk.sampler.MetropolisLocal(machine=ma)

    op = nk.optimizer.Sgd(ma, learning_rate=0.1)

    vmc = nk.Vmc(ha, sa, op, n_samples=n_samples, **kwargs)

    # Add custom observable
    X = [[0, 1], [1, 0]]
    sx = nk.operator.LocalOperator(hi, [X] * 8, [[i] for i in range(8)])

    return ma, vmc, sx


def test_vmc_advance():
    ma1, vmc1, _ = _setup_vmc(n_samples=500, diag_shift=0.01)
    for i in range(10):
        vmc1.advance()

    ma2, vmc2, _ = _setup_vmc(n_samples=500, diag_shift=0.01)
    for step in vmc2.iter(10):
        pass

    assert (ma1.parameters == ma2.parameters).all()


def test_vmc_advance_iterative():
    ma1, vmc1, _ = _setup_vmc(n_samples=500, diag_shift=0.01, use_iterative=True)
    for i in range(10):
        vmc1.advance()

    ma2, vmc2, _ = _setup_vmc(n_samples=500, diag_shift=0.01, use_iterative=True)
    for step in vmc2.iter(10):
        pass

    assert (ma1.parameters == ma2.parameters).all()


def test_vmc_iterator():
    ma, vmc, sx = _setup_vmc(n_samples=500, diag_shift=0.01)
    operators = {"Energy": vmc._ham, "SigmaX": sx}

    count = 0
    last_obs = None
    for i, step in enumerate(vmc.iter(300)):
        count += 1
        assert step == i
        obs = vmc.estimate(operators)
        for name in "Energy", "SigmaX":
            assert name in obs
            e = obs[name]
            assert (
                hasattr(e, "mean")
                and hasattr(e, "error_of_mean")
                and hasattr(e, "variance")
                and hasattr(e, "tau_corr")
                and hasattr(e, "R_hat")
            )
        last_obs = obs

    assert count == 300
    assert last_obs["Energy"].mean == approx(-10.25, abs=0.2)


def test_vmc_iterator_iterative():
    ma, vmc, sx = _setup_vmc(n_samples=500, diag_shift=0.01, use_iterative=True)
    operators = {"Energy": vmc._ham, "SigmaX": sx}

    count = 0
    last_obs = None
    for i, step in enumerate(vmc.iter(300)):
        count += 1
        assert step == i
        obs = vmc.estimate(operators)
        # TODO: Choose which version we want
        # for name in "Energy", "EnergyVariance", "SigmaX":
        #     assert name in obs
        #     e = obs[name]
        #     assert "Mean" in e and "Sigma" in e and "Taucorr" in e
        for name in "Energy", "SigmaX":
            assert name in obs
            e = obs[name]
            assert hasattr(e, "mean") and hasattr(e, "variance") and hasattr(e, "R_hat")
        last_obs = obs

    assert count == 300
    assert last_obs["Energy"].mean == approx(-10.25, abs=0.2)


def test_vmc_run():
    ma, vmc, sx = _setup_vmc(n_samples=500, diag_shift=0.01)
    obs = {"SigmaX": sx}

    tempdir = tempfile.mkdtemp()
    print("Writing test output files to: {}".format(tempdir))
    prefix = tempdir + "/vmc_test"
    vmc.run(300, prefix, obs)

    with open(prefix + ".log") as logfile:
        log = json.load(logfile)

    shutil.rmtree(tempdir)

    assert "Output" in log
    output = log["Output"]
    assert len(output) == 300

    for i, obs in enumerate(output):
        step = obs["Iteration"]
        assert step == i
        for name in "Energy", "SigmaX":
            assert name in obs
            e = obs[name]
            assert "Mean" in e
            assert "Sigma" in e
            assert "TauCorr" in e
        last_obs = obs

    assert last_obs["Energy"]["Mean"] == approx(-10.25, abs=0.2)


def test_imag_time_propagation():
    L = 8
    g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)
    hi = nk.hilbert.Spin(s=0.5, N=L)
    ha = nk.operator.Ising(graph=g, h=0.0, hilbert=hi)

    psi0 = np.random.rand(hi.n_states)
    driver = nk.exact.PyExactTimePropagation(
        ha,
        t0=0,
        dt=0.1,
        initial_state=psi0,
        propagation_type="imaginary",
        solver_kwargs={"atol": 1e-10, "rtol": 1e-10},
    )

    driver.advance(1000)
    assert driver.estimate(ha).mean.real == approx(-8.0)



from ..abstract_sampler import AbstractSampler

import jax
from functools import partial
from netket.legacy import random as _random


class MetropolisHastings(AbstractSampler):
    def __init__(self, machine, kernel, n_chains=16, sweep_size=None, rng_key=None):

        super().__init__(machine, n_chains)

        self._random_state_kernel = jax.jit(kernel.random_state)
        self._transition_kernel = jax.jit(kernel.transition)

        self._rng_key = rng_key
        if rng_key is None:
            self._rng_key = jax.random.PRNGKey(
                _random.randint(low=0, high=2 ** 32, size=()).item()
            )

        self.machine_pow = 2

        self.n_chains = n_chains

        self.sweep_size = sweep_size

    @staticmethod
    @partial(jax.jit, static_argnums=(0, 1, 2))
    def _metropolis_kernel(
        logpdf,
        transition_kernel,
        n_samples,
        sweep_size,
        initial_state,
        params,
        machine_pow,
        rng_key,
    ):
        # Array shapes are effectively static in jax-jitted code, so code will be
        # recompiled every time this changes.
        n_chains = initial_state.shape[0]

        def chains_one_step(i, walker):
            key, state, log_prob = walker

            # 1 to propagate for next iteration, 1 for uniform rng and n_chains for transition kernel
            keys = jax.random.split(key, 2 + n_chains)

            proposal = jax.vmap(transition_kernel, in_axes=(0, 0), out_axes=0)(
                keys[2:], state
            )
            proposal_log_prob = machine_pow * logpdf(params, proposal).real

            uniform = jax.random.uniform(keys[1], shape=(n_chains,))
            do_accept = uniform < jax.numpy.exp(proposal_log_prob - log_prob)

            # do_accept must match ndim of proposal and state (which is 2)
            state = jax.numpy.where(do_accept.reshape(-1, 1), proposal, state)

            log_prob = jax.numpy.where(
                do_accept.reshape(-1), proposal_log_prob, log_prob
            )

            return (keys[0], state, log_prob)

        # Loop over the sweeps
        def chains_one_sweep(walker, i):
            key, state, log_prob = walker
            walker = jax.lax.fori_loop(
                0, sweep_size, chains_one_step, (key, state, log_prob)
            )
            return walker, walker[1]

        keys = jax.random.split(rng_key, 2)
        initial_log_prob = machine_pow * logpdf(params, initial_state).real

        # Loop over the samples
        walker, samples = jax.lax.scan(
            chains_one_sweep,
            (keys[1], initial_state, initial_log_prob),
            xs=None,
            length=n_samples,
        )

        return keys[0], samples

    @property
    def n_chains(self):
        return self._n_chains

    @n_chains.setter
    def n_chains(self, n_chains):
        if n_chains < 0:
            raise ValueError("Expected a positive integer for n_chains ")

        self._n_chains = int(n_chains)
        self.reset(True)

    @property
    def machine_pow(self):
        return self._machine_pow

    @machine_pow.setter
    def machine_pow(self, m_power):
        self._machine_pow = m_power

    @property
    def sweep_size(self):
        return self._sweep_size

    @sweep_size.setter
    def sweep_size(self, sweep_size):
        self._sweep_size = sweep_size if sweep_size != None else self._input_size
        if self._sweep_size < 0:
            raise ValueError("Expected a positive integer for sweep_size ")

    def reset(self, init_random=False):
        if init_random:

            self._rng_key, self._state = jax.lax.scan(
                self._random_state_kernel, self._rng_key, xs=None, length=self._n_chains
            )

            assert self._state.shape == self.sample_shape

        self._accepted_samples = 0
        self._total_samples = 0

    def generate_samples(self, n_samples, init_random=False, samples=None):
        if n_samples == 0:
            return

        self.reset(init_random)

        self._rng_key, samples = self._metropolis_kernel(
            self.machine._forward_fn_nj,
            self._transition_kernel,
            n_samples,
            self.sweep_size,
            self._state,
            self.machine.parameters,
            self.machine_pow,
            self._rng_key,
        )

        self._state = samples[-1]
        return samples

    def __next__(self):
        self._rng_key, samples = self._metropolis_kernel(
            self.machine._forward_fn_nj,
            self._transition_kernel,
            1,
            self.sweep_size,
            self._state,
            self.machine.parameters,
            self.machine_pow,
            self._rng_key,
        )

        self._state = samples[-1]
        return self._state



import pytest

import json
from pytest import approx
import netket.legacy as nk
import numpy as np
import shutil
import tempfile

pytestmark = pytest.mark.legacy

SEED = 3141592
L = 4

sx = [[0, 1], [1, 0]]
sy = [[0, -1j], [1j, 0]]
sz = [[1, 0], [0, -1]]

sigmam = [[0, 0], [1, 0]]


def _setup_system():
    hi = nk.hilbert.Spin(s=0.5) ** L

    ha = nk.operator.LocalOperator(hi)
    j_ops = []
    for i in range(L):
        ha += (0.3 / 2.0) * nk.operator.LocalOperator(hi, sx, [i])
        ha += (2.0 / 4.0) * nk.operator.LocalOperator(
            hi, np.kron(sz, sz), [i, (i + 1) % L]
        )
        j_ops.append(nk.operator.LocalOperator(hi, sigmam, [i]))

    # Create the liouvillian
    lind = nk.operator.LocalLiouvillian(ha, j_ops)
    return hi, lind


def _setup_ss(**kwargs):
    nk.random.seed(SEED)
    np.random.seed(SEED)

    hi, lind = _setup_system()

    ma = nk.machine.density_matrix.RbmSpin(hilbert=hi, alpha=1)
    ma.init_random_parameters(sigma=0.01, seed=SEED)

    sa = nk.sampler.MetropolisLocal(machine=ma)
    sa_obs = nk.sampler.MetropolisLocal(machine=ma.diagonal())

    op = nk.optimizer.Sgd(ma, learning_rate=0.1)

    if "sr" in kwargs:
        sr = nk.optimizer.SR(ma, **kwargs["sr"])
        kwargs["sr"] = sr

    ss = nk.SteadyState(
        lindblad=lind, sampler=sa, optimizer=op, sampler_obs=sa_obs, **kwargs
    )

    return ma, ss


def _setup_obs():
    hi = nk.hilbert.Spin(s=0.5) ** L

    obs_sx = nk.operator.LocalOperator(hi)
    for i in range(L):
        obs_sx += nk.operator.LocalOperator(hi, sx, [i])

    obs = {"SigmaX": obs_sx}
    return obs


def test_ss_advance():
    ma1, vmc1 = _setup_ss(n_samples=500, n_samples_obs=250)
    for i in range(10):
        vmc1.advance()

    ma2, vmc2 = _setup_ss(n_samples=500, n_samples_obs=250)
    for step in vmc2.iter(10):
        pass

    assert (ma1.parameters == ma2.parameters).all()


def test_ss_advance_sr():
    sr = {"diag_shift": 0.01, "use_iterative": False}

    ma1, vmc1 = _setup_ss(n_samples=500, n_samples_obs=250, sr=sr)
    for i in range(10):
        vmc1.advance()

    sr = {"diag_shift": 0.01, "use_iterative": False}
    ma2, vmc2 = _setup_ss(n_samples=500, n_samples_obs=250, sr=sr)
    for step in vmc2.iter(10):
        pass

    assert (ma1.parameters == ma2.parameters).all()


def test_ss_advance_sr_iterative():
    sr = {"diag_shift": 0.01, "use_iterative": True}
    ma1, vmc1 = _setup_ss(n_samples=500, n_samples_obs=250, sr=sr)
    for i in range(10):
        vmc1.advance()

    sr = {"diag_shift": 0.01, "use_iterative": True}
    ma2, vmc2 = _setup_ss(n_samples=500, n_samples_obs=250, sr=sr)
    for step in vmc2.iter(10):
        pass

    assert (ma1.parameters == ma2.parameters).all()


def test_ss_iterator():
    ma, vmc = _setup_ss(n_samples=800, n_samples_obs=250)
    obs_op = _setup_obs()

    N_iters = 160
    count = 0
    losses = []
    for i, step in enumerate(vmc.iter(N_iters)):
        count += 1
        assert step == i
        obs = vmc.estimate(obs_op)
        obs["LdagL"] = vmc.ldagl
        for name in "LdagL", "SigmaX":
            assert name in obs
            e = obs[name]
            assert (
                hasattr(e, "mean")
                and hasattr(e, "error_of_mean")
                and hasattr(e, "variance")
                and hasattr(e, "tau_corr")
                and hasattr(e, "R_hat")
            )
        losses.append(vmc.ldagl["mean"])

    assert count == N_iters
    assert np.mean(losses[-10:]) == approx(0.0, abs=0.003)


def test_ss_iterator_sr():
    sr = {"diag_shift": 0.01, "use_iterative": False}
    ma, vmc = _setup_ss(n_samples=800, n_samples_obs=250, sr=sr)
    obs_op = _setup_obs()

    N_iters = 160
    count = 0
    losses = []
    for i, step in enumerate(vmc.iter(N_iters)):
        count += 1
        assert step == i
        obs = vmc.estimate(obs_op)
        obs["LdagL"] = vmc.ldagl
        # TODO: Choose which version we want
        # for name in "Energy", "EnergyVariance", "SigmaX":
        #     assert name in obs
        #     e = obs[name]
        #     assert "Mean" in e and "Sigma" in e and "Taucorr" in e
        for name in "LdagL", "SigmaX":
            assert name in obs
            e = obs[name]
            assert hasattr(e, "mean") and hasattr(e, "variance") and hasattr(e, "R_hat")
        losses.append(vmc.ldagl["mean"])

    assert count == N_iters
    # Should readd, but pure python machines are not good enough to converge
    # requires jax. split into different file?
    # assert np.mean(losses[-10:]) == approx(0.0, abs=0.003)


def test_ss_run():
    ma, vmc = _setup_ss(n_samples=1000, n_samples_obs=250)
    obs_op = _setup_obs()

    N_iters = 200

    tempdir = tempfile.mkdtemp()
    print("Writing test output files to: {}".format(tempdir))
    prefix = tempdir + "/ss_test"
    vmc.run(N_iters, prefix, obs=obs_op)

    with open(prefix + ".log") as logfile:
        log = json.load(logfile)

    shutil.rmtree(tempdir)

    assert "Output" in log
    output = log["Output"]
    assert len(output) == N_iters

    losses = []
    for i, obs in enumerate(output):
        step = obs["Iteration"]
        assert step == i
        for name in "LdagL", "SigmaX":
            assert name in obs
            e = obs[name]
            assert "Mean" in e
            assert "Sigma" in e
            assert "TauCorr" in e
        losses.append(vmc.ldagl["Mean"])

    # Should readd, but pure python machines are not good enough to converge
    # requires jax. split into different file?
    # assert np.mean(losses[-10:]) == approx(0.0, abs=0.003)


def _ldagl(par, machine, L):
    machine.parameters = np.copy(par)
    rho = machine.to_array()
    return np.sum(np.abs(L @ rho) ** 2)


def central_diff_grad(func, x, eps, *args):
    grad = np.zeros(len(x), dtype=complex)
    epsd = np.zeros(len(x), dtype=complex)
    epsd[0] = eps
    for i in range(len(x)):
        assert not np.any(np.isnan(x + epsd))
        grad_r = 0.5 * (func(x + epsd, *args) - func(x - epsd, *args))
        grad_i = 0.5 * (func(x + 1j * epsd, *args) - func(x - 1j * epsd, *args))
        grad[i] = 0.5 * grad_r + 0.5j * grad_i
        assert not np.isnan(grad[i])
        grad[i] /= eps
        epsd = np.roll(epsd, 1)
    return grad


def same_derivatives(der_log, num_der_log, eps=1.0e-6):
    assert np.max(np.real(der_log - num_der_log)) == approx(0.0, rel=eps, abs=eps)
    # The imaginary part is a bit more tricky, there might be an arbitrary phase shift
    assert np.max(
        np.abs(np.exp(np.imag(der_log - num_der_log) * 1.0j) - 1.0)
    ) == approx(0.0, rel=eps, abs=eps)


# disable test because to make it work we need MUCH more samples than travis
# can handle
def _test_ldagl_gradient():
    ma, driver = _setup_ss(n_samples=500, n_samples_obs=250)
    lind = driver._lind
    pars = np.copy(ma.parameters)

    driver._sr = None
    grad_exact = central_diff_grad(_ldagl, pars, 1.0e-5, ma, lind.to_sparse())

    driver.n_samples = 2e5
    driver.n_discard = 5e2
    driver.machine.parameters = np.copy(pars)
    grad_approx = driver._forward_and_backward()

    err = 5 / np.sqrt(driver.n_samples)
    same_derivatives(np.real(grad_approx), np.real(grad_exact), eps=err)



import netket.legacy as nk
import networkx as nx
import numpy as np
import pytest
from pytest import approx
import os
from netket.hilbert import Spin

import pytest

pytestmark = pytest.mark.legacy

test_jax = nk.utils.jax_available
try:
    import torch

    test_torch = True
except:
    test_torch = False


def merge_dicts(x, y):
    z = x.copy()  # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z


machines = {}
dm_machines = {}

# TESTS FOR SPIN HILBERT
# Constructing a 1d lattice
g = nk.graph.Hypercube(length=4, n_dim=1)

# Hilbert space of spins from given graph
hi = Spin(s=0.5, N=g.n_nodes)


if test_jax:
    import jax
    import jax.experimental
    import jax.experimental.stax

    def initializer(rng, shape):
        return np.random.normal(scale=0.05, size=shape)

    # machines["Jax Real"] = nk.machine.Jax(
    #     hi,
    #     jax.experimental.stax.serial(
    #         jax.experimental.stax.Dense(4, initializer, initializer),
    #         jax.experimental.stax.Relu,
    #         jax.experimental.stax.Dense(2, initializer, initializer),
    #         jax.experimental.stax.Relu,
    #         jax.experimental.stax.Dense(2, initializer, initializer),
    #     ),
    #     dtype=float,
    # )

    machines["Jax Complex"] = nk.machine.Jax(
        hi,
        jax.experimental.stax.serial(
            jax.experimental.stax.Dense(4, initializer, initializer),
            jax.experimental.stax.Tanh,
            jax.experimental.stax.Dense(2, initializer, initializer),
            jax.experimental.stax.Tanh,
            jax.experimental.stax.Dense(1, initializer, initializer),
        ),
        dtype=complex,
    )

    machines["Jax mps"] = nk.machine.MPSPeriodic(hi, graph=g, bond_dim=4, dtype=complex)

    machines["Jax RbmSpinPhase (R->C)"] = nk.machine.JaxRbmSpinPhase(hi, alpha=1)

    dm_machines["Jax NDM"] = nk.machine.density_matrix.NdmSpinPhase(hi, alpha=1, beta=1)


if test_torch:
    import torch

    input_size = hi.size
    alpha = 1

    model = torch.nn.Sequential(
        torch.nn.Linear(input_size, alpha * input_size),
        torch.nn.Sigmoid(),
        torch.nn.Linear(alpha * input_size, 2),
        torch.nn.Sigmoid(),
    )
    machines["Torch"] = nk.machine.Torch(model, hilbert=hi)


machines["RbmSpin 1d Hypercube spin"] = nk.machine.RbmSpin(hilbert=hi, alpha=2)

machines["RbmSpinSymm 1d Hypercube spin"] = nk.machine.RbmSpinSymm(
    hilbert=hi, alpha=2, automorphisms=g
)

machines["Real RBM"] = nk.machine.RbmSpinReal(hilbert=hi, alpha=2)

machines["Phase RBM"] = nk.machine.RbmSpinPhase(hilbert=hi, alpha=2)

machines["Jastrow 1d Hypercube spin"] = nk.machine.Jastrow(hilbert=hi)

hi = Spin(s=0.5, N=g.n_nodes, total_sz=0)
machines["Jastrow 1d Hypercube spin symm bias"] = nk.machine.Jastrow(
    hilbert=hi, use_visible_bias=True, automorphisms=g
)

dm_machines["Phase NDM"] = nk.machine.density_matrix.RbmSpin(
    hilbert=hi,
    alpha=2,
    use_visible_bias=True,
    use_hidden_bias=True,
)


# machines["MPS Diagonal 1d spin"] = nk.machine.MPSPeriodicDiagonal(hi, bond_dim=3)
# machines["MPS 1d spin"] = nk.machine.MPSPeriodic(hi, bond_dim=3)

# BOSONS
hi = nk.hilbert.Boson(N=g.n_nodes, n_max=3)
machines["RbmSpin 1d Hypercube boson"] = nk.machine.RbmSpin(hilbert=hi, alpha=1)

machines["RbmSpinSymm 1d Hypercube boson"] = nk.machine.RbmSpinSymm(
    hilbert=hi, alpha=2, automorphisms=g
)
machines["RbmMultiVal 1d Hypercube boson"] = nk.machine.RbmMultiVal(
    hilbert=hi, n_hidden=2
)
machines["Jastrow 1d Hypercube boson"] = nk.machine.Jastrow(hilbert=hi)

machines["JastrowSymm 1d Hypercube boson real"] = nk.machine.JastrowSymm(
    hilbert=hi, dtype=float, automorphisms=g
)
# machines["MPS 1d boson"] = nk.machine.MPSPeriodic(hi, bond_dim=4)


np.random.seed(12346)


def same_derivatives(der_log, num_der_log, eps=1.0e-5):
    assert der_log.shape == num_der_log.shape
    # assert np.max(np.real(der_log - num_der_log)) == approx(0.0, rel=eps, abs=eps)

    np.testing.assert_array_almost_equal(der_log.real, num_der_log.real, decimal=4)

    # The imaginary part is a bit more tricky, there might be an arbitrary phase shift
    assert np.max(
        np.abs(np.exp(np.imag(der_log - num_der_log) * 1.0j) - 1.0)
    ) == approx(0.0, rel=eps, abs=eps)


def log_val_f(par, machine, v):
    machine.parameters = machine.numpy_unflatten(par, machine.parameters)
    if v.ndim != 1:
        if v.size != v.shape[1]:
            raise RuntimeError(
                "numerical derivatives can be tested only for non batched inputs"
            )

    return machine.log_val(v.reshape(1, -1))[0]


def log_val_vec_f(par, machine, v, vec):
    machine.parameters = machine.numpy_unflatten(par, machine.parameters)
    if v.ndim == 2:
        out_val = machine.log_val(v)
    else:
        out_val = machine.log_val(v.reshape(1, -1))
    assert vec.shape == out_val.shape
    return np.vdot(out_val, vec)


def central_diff_grad(func, x, eps, *args):
    grad = np.zeros(len(x), dtype=complex)
    epsd = np.zeros(len(x), dtype=complex)
    epsd[0] = eps
    for i in range(len(x)):
        assert not np.any(np.isnan(x + epsd))
        grad[i] = 0.5 * (func(x + epsd, *args) - func(x - epsd, *args))
        assert not np.isnan(grad[i])
        grad[i] /= eps
        epsd = np.roll(epsd, 1)
    return grad


def test_set_get_parameters():
    for name, machine in merge_dicts(machines, dm_machines).items():
        unflatten = machine.numpy_unflatten
        flatten = machine.numpy_flatten

        print("Machine test: %s" % name)
        assert machine.n_par > 0
        npar = machine.n_par
        machine.init_random_parameters()
        randpars = flatten(machine.parameters)

        if machine.has_complex_parameters:
            assert np.array_equal(flatten(machine.parameters), randpars)
            assert not all(randpars.real == 0)
            assert not all(randpars.imag == 0)
        else:
            assert np.array_equal(flatten(machine.parameters).real, randpars.real)
            assert not all(randpars.real == 0)

        machine.parameters = unflatten(np.zeros(npar), machine.parameters)
        assert np.count_nonzero(np.abs(flatten(machine.parameters))) == 0


def test_save_load_parameters(tmpdir):
    for name, machine in merge_dicts(machines, dm_machines).items():
        print("Machine test: %s" % name)
        assert machine.n_par > 0
        n_par = machine.n_par

        unflatten = machine.numpy_unflatten
        flatten = machine.numpy_flatten

        machine.init_random_parameters()
        randpars = flatten(machine.parameters)

        machine.parameters = unflatten(randpars, machine.parameters)
        fn = tmpdir.mkdir("datawf").join("test.wf")

        filename = os.path.join(fn.dirname, fn.basename)

        machine.save(filename)
        machine.parameters = unflatten(np.zeros(n_par), machine.parameters)
        machine.load(filename)

        os.remove(filename)
        os.rmdir(fn.dirname)
        if machine.has_complex_parameters:
            assert np.array_equal(flatten(machine.parameters), randpars)
        else:
            assert np.array_equal(flatten(machine.parameters).real, randpars.real)


def test_log_derivative():
    np.random.seed(12345)

    for name, machine in merge_dicts(machines, dm_machines).items():

        print("Machine test: %s" % name)

        npar = machine.n_par

        # random visibile state
        hi = machine.hilbert
        assert hi.size > 0

        v = np.zeros(machine.input_size)

        flatten = machine.numpy_flatten

        for i in range(100):
            if name in dm_machines:
                hi.random_state(out=v[: hi.size])
                hi.random_state(out=v[hi.size :])
            else:
                hi.random_state(out=v)

            machine.init_random_parameters(seed=i)
            randpars = flatten(machine.parameters)

            der_log = flatten(machine.der_log(v.reshape((1, -1))))

            assert der_log.shape == (machine.n_par,)

            if "Jastrow" in name:
                assert np.max(np.imag(der_log)) == approx(0.0)

            num_der_log = central_diff_grad(log_val_f, randpars, 1.0e-9, machine, v)

            same_derivatives(der_log, num_der_log)
            # print(np.linalg.norm(der_log - num_der_log))
            # Check if machine is correctly set to be holomorphic
            # The check is done only on smaller subset of parameters, for speed


def test_vector_jacobian():
    for name, machine in merge_dicts(machines, dm_machines).items():
        print("Machine test: %s" % name)

        npar = machine.n_par

        # random visibile state
        hi = machine.hilbert
        if name in dm_machines:
            hi = nk.hilbert.DoubledHilbert(hi)
        assert hi.size > 0

        batch_size = 100
        v = np.zeros((batch_size, machine.input_size))

        flatten = machine.numpy_flatten

        for i in range(batch_size):
            hi.random_state(out=v[i])

        machine.init_random_parameters(seed=1234, sigma=0.1)
        randpars = flatten(machine.parameters)

        vec = np.random.uniform(size=batch_size) + 1.0j * np.random.uniform(
            size=batch_size
        ) / float(batch_size)

        vjp = machine.vector_jacobian_prod(v, vec)
        vjp = flatten(vjp)

        num_der_log = central_diff_grad(
            log_val_vec_f, randpars, 1.0e-6, machine, v, vec
        )

        same_derivatives(vjp, num_der_log)


def test_input_size():
    for name, machine in machines.items():
        print("Machine test: %s" % name)
        hi = machine.hilbert

        assert machine.input_size == hi.size

    for name, machine in dm_machines.items():
        print("Machine test: %s" % name)
        hi = machine.hilbert

        assert machine.input_size == 2 * hi.size


def test_to_array():
    for name, machine in merge_dicts(machines, dm_machines).items():
        print("Machine test: %s" % name)

        npar = machine.n_par

        randpars = 0.5 * (np.random.randn(npar) + 1.0j * np.random.randn(npar))
        if "Torch" in name:
            randpars = randpars.real
        machine.parameters = machine.numpy_unflatten(randpars, machine.parameters)

        hi = machine.hilbert
        if name in dm_machines:
            hi = nk.hilbert.DoubledHilbert(hi)

        all_psis = machine.to_array(normalize=False)
        # test shape
        assert all_psis.shape[0] == hi.n_states
        assert len(all_psis.shape) == 1

        log_vals = machine.log_val(hi.all_states())
        logmax = log_vals.real.max()

        norm = (np.abs(np.exp(log_vals - logmax)) ** 2).sum()

        # Tests broken for dm machines because of norm. should split
        if name in dm_machines:
            continue

        # test random values
        for i in range(100):
            rstate = np.zeros(hi.size)
            local_states = hi.local_states
            hi.random_state(out=rstate)

            number = hi.states_to_numbers(rstate)

            assert np.abs(
                np.exp(machine.log_val(rstate.reshape(1, -1)) - logmax)
                - all_psis[number]
            ) == approx(0.0)

        # test rescale
        all_psis_normalized = machine.to_array(normalize=True)

        # test random values
        for i in range(100):
            rstate = np.zeros(hi.size)
            local_states = hi.local_states
            hi.random_state(out=rstate)

            number = hi.states_to_numbers(rstate)

            assert np.abs(
                np.exp(machine.log_val(rstate.reshape(1, -1)) - logmax) / np.sqrt(norm)
                - all_psis_normalized[number]
            ) == approx(0.0)



import json as _json
from os import path as _path
from netket.legacy.vmc_common import tree_map as _tree_map


def _exists_json(prefix):
    return _path.exists(prefix + ".log") or _path.exists(prefix + ".wf")


def _to_json(ob):
    to_json = getattr(ob, "to_json", None)
    if to_json is not None:
        return ob.to_json()
    else:
        return ob


class JsonLog:
    """
    Creates a Json Logger sink object, that can be passed with keyword argument `logger` to Monte
    Carlo drivers in order to serialize the outpit data of the simulation.

    Args:
        output_prefix: the name of the output files before the extension
        save_params_every: every how many iterations should machine parameters be flushed to file
        write_every: every how many iterations should data be flushed to file
        mode: Specify the behaviour in case the file already exists at this output_prefix. Options
        are
        - `[w]rite`: (default) overwrites file if it already exists;
        - `[a]ppend`: appends to the file if it exists, overwise creates a new file;
        - `[x]` or `fail`: fails if file already exists;
    """

    def __init__(
        self, output_prefix, mode="write", save_params_every=50, write_every=50
    ):
        # Shorthands for mode
        if mode == "w":
            mode = "write"
        elif mode == "a":
            mode = "append"
        elif mode == "x":
            mode = "fail"

        if not ((mode == "write") or (mode == "append") or (mode == "fail")):
            raise ValueError(
                "Mode not recognized: should be one of `[w]rite`, `[a]ppend` or `[x]`(fail)."
            )

        file_exists = _exists_json(output_prefix)

        starting_json_content = {"Output": []}

        if file_exists and mode == "append":
            # if there is only the .wf file but not the json one, raise an error
            if not _path.exists(output_prefix + ".log"):
                raise ValueError(
                    "History file does not exists, but wavefunction file does. Please change `output_prefix or set mode=`write`."
                )

            starting_json_content = _json.load(open(output_prefix + ".log"))

        elif file_exists and mode == "fail":
            raise ValueError(
                "Output file already exists. Either delete it manually or change `output_prefix`."
            )

        self._json_out = starting_json_content
        self._prefix = output_prefix
        self._write_every = write_every
        self._save_params_every = save_params_every
        self._old_step = 0
        self._steps_notflushed_write = 0
        self._steps_notflushed_pars = 0

    def __call__(self, step, item, machine):
        item["Iteration"] = step

        self._json_out["Output"].append(item)

        if (
            self._steps_notflushed_write % self._write_every == 0
            or step == self._old_step - 1
        ):
            self._flush_log()
        if (
            self._steps_notflushed_pars % self._save_params_every == 0
            or step == self._old_step - 1
        ):
            self._flush_params(machine)

        self._old_step = step
        self._steps_notflushed_write += 1
        self._steps_notflushed_pars += 1

    def _flush_log(self):
        with open(self._prefix + ".log", "w") as outfile:
            log_data = _tree_map(_to_json, self._json_out)
            _json.dump(log_data, outfile)
            self._steps_notflushed_write = 0

    def _flush_params(self, machine):
        machine.save(self._prefix + ".wf")
        self._steps_notflushed_pars = 0

    def flush(self, machine=None):
        """
        Writes to file the content of this logger.

        :param machine: optionally also writes the parameters of the machine.
        """
        self._flush_log()

        if machine is not None:
            self._flush_params(machine)



from .metropolis_hastings import *
from ._kernels import _HamiltonianKernel


def MetropolisHamiltonian(machine, hamiltonian, n_chains=16, sweep_size=None, **kwargs):
    r"""
    Sampling based on the off-diagonal elements of a Hamiltonian (or a generic Operator).
    In this case, the transition matrix is taken to be:

    .. math::
       T( \mathbf{s} \rightarrow \mathbf{s}^\prime) = \frac{1}{\mathcal{N}(\mathbf{s})}\theta(|H_{\mathbf{s},\mathbf{s}^\prime}|),

    where :math:`\theta(x)` is the Heaviside step function, and :math:`\mathcal{N}(\mathbf{s})`
    is a state-dependent normalization.
    The effect of this transition probability is then to connect (with uniform probability)
    a given state :math:`\mathbf{s}` to all those states :math:`\mathbf{s}^\prime` for which the Hamiltonian has
    finite matrix elements.
    Notice that this sampler preserves by construction all the symmetries
    of the Hamiltonian. This is in generally not true for the local samplers instead.

    Args:
       machine: A machine :math:`\Psi(s)` used for the sampling.
                The probability distribution being sampled
                from is :math:`F(\Psi(s))`, where the function
                :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
       hamiltonian: The operator used to perform off-diagonal transition.
       n_chains: The number of Markov Chain to be run in parallel on a single process.
       sweep_size: The number of exchanges that compose a single sweep.
                   If None, sweep_size is equal to the number of degrees of freedom (n_visible).


    Examples:
       Sampling from a RBM machine in a 1D lattice of spin 1/2

       >>> import netket as nk
       >>>
       >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
       >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
       >>>
       >>> # RBM Spin Machine
       >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
       >>>
       >>> # Transverse-field Ising Hamiltonian
       >>> ha = nk.operator.Ising(hilbert=hi, h=1.0)
       >>>
       >>> # Construct a MetropolisHamiltonian Sampler
       >>> sa = nk.sampler.MetropolisHamiltonian(machine=ma,hamiltonian=ha)
    """
    return MetropolisHastings(
        machine,
        _HamiltonianKernel(machine, hamiltonian),
        n_chains,
        sweep_size,
        **kwargs,
    )


def MetropolisHamiltonianPt(
    machine, hamiltonian, n_replicas=16, sweep_size=None, **kwargs
):
    r"""
    This sampler performs parallel-tempering
    moves in addition to the local moves implemented in `MetropolisLocal`.
    The number of replicas can be :math:`N_{\mathrm{rep}}` chosen by the user.

    Args:
        machine: A machine :math:`\Psi(s)` used for the sampling.
                  The probability distribution being sampled
                  from is :math:`F(\Psi(s))`, where the function
                  :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
        hamiltonian: The operator used to perform off-diagonal transition.
        n_replicas: The number of replicas used for parallel tempering.
        sweep_size: The number of exchanges that compose a single sweep.
                     If None, sweep_size is equal to the number of degrees of freedom (n_visible).
        batch_size: The batch size to be used when calling log_val on the given Machine.
                    If None, batch_size is equal to the number of replicas (n_replicas).
    """
    return MetropolisHastingsPt(
        machine,
        _HamiltonianKernel(machine, hamiltonian),
        n_replicas,
        sweep_size,
        **kwargs,
    )



from functools import singledispatch
from . import numpy


@singledispatch
def AdaDelta(machine, rho=0.95, epscut=1.0e-7, l2reg=0):
    r"""AdaDelta Optimizer.
        Like RMSProp, `AdaDelta <http://arxiv.org/abs/1212.5701>`_ corrects the
        monotonic decay of learning rates associated with AdaGrad,
        while additionally eliminating the need to choose a global
        learning rate :math:`\eta`. The NetKet naming convention of
        the parameters strictly follows the one introduced in the original paper;
        here :math:`E[g^2]` is equivalent to the vector :math:`\mathbf{s}` from RMSProp.
        :math:`E[g^2]` and :math:`E[\Delta x^2]` are initialized as zero vectors.

        .. math:: E[g^2]^\prime_k &= \rho E[g^2] + (1-\rho)G_k(\mathbf{p})^2\\
                  \Delta p_k &= - \frac{\sqrt{E[\Delta x^2]+\epsilon}}{\sqrt{E[g^2]+ \epsilon}}G_k(\mathbf{p})\\
                  E[\Delta x^2]^\prime_k &= \rho E[\Delta x^2] + (1-\rho)\Delta p_k^2\\
                  p^\prime_k &= p_k + \Delta p_k


        Args:
           rho: Exponential decay rate, in [0,1].
           epscut: Small :math:`\epsilon` cutoff.
           l2reg (float): The amount of L2 regularization.

        Examples:
           Simple AdaDelta optimizer.

           >>> from netket.optimizer import AdaDelta
           >>> op = AdaDelta()
    """
    return numpy.AdaDelta(rho, epscut, l2reg)



from functools import singledispatch
import numpy as _np
from . import numpy

# Register numpy kernels here
@singledispatch
def _LocalKernel(machine):
    return numpy._LocalKernel(
        _np.asarray(machine.hilbert.local_states), machine.input_size
    )


@singledispatch
def _ExchangeKernel(machine, clusters):
    return numpy._ExchangeKernel(machine.hilbert, clusters)


@singledispatch
def _CustomKernel(machine, move_operators, move_weights=None):
    return numpy._CustomKernel(move_operators, move_weights)


@singledispatch
def _HamiltonianKernel(machine, hamiltonian):
    return numpy._HamiltonianKernel(hamiltonian)



from pytest import approx
import netket as nk
import numpy as np

from .. import common

pytestmark = common.skipif_mpi

SEED = 3141592
L = 4

sx = [[0, 1], [1, 0]]
sy = [[0, -1j], [1j, 0]]
sz = [[1, 0], [0, -1]]

sigmam = [[0, 0], [1, 0]]


def _setup_system():
    hi = nk.hilbert.Spin(s=0.5) ** L

    ha = nk.operator.LocalOperator(hi)
    j_ops = []
    for i in range(L):
        ha += (0.3 / 2.0) * nk.operator.LocalOperator(hi, sx, [i])
        ha += (2.0 / 4.0) * nk.operator.LocalOperator(
            hi, np.kron(sz, sz), [i, (i + 1) % L]
        )
        j_ops.append(nk.operator.LocalOperator(hi, sigmam, [i]))

    # Create the liouvillian
    lind = nk.operator.LocalLiouvillian(ha, j_ops)
    return hi, lind


def _setup_ss(**kwargs):
    nk.random.seed(SEED)
    np.random.seed(SEED)

    hi, lind = _setup_system()

    ma = nk.machine.density_matrix.RbmSpin(hilbert=hi, alpha=1)
    ma.init_random_parameters(sigma=0.01, seed=SEED)

    sa = nk.sampler.MetropolisLocal(machine=ma)
    sa_obs = nk.sampler.MetropolisLocal(machine=ma.diagonal())

    op = nk.optimizer.Sgd(ma, learning_rate=0.1)

    if "sr" in kwargs:
        sr = nk.optimizer.SR(ma, **kwargs["sr"])
        kwargs["sr"] = sr

    ss = nk.SteadyState(
        lindblad=lind, sampler=sa, optimizer=op, sampler_obs=sa_obs, **kwargs
    )

    return ma, ss


def _setup_obs():
    hi = nk.hilbert.Spin(s=0.5) ** L

    obs_sx = nk.operator.LocalOperator(hi)
    for i in range(L):
        obs_sx += nk.operator.LocalOperator(hi, sx, [i])

    obs = {"SigmaX": obs_sx}
    return obs


def test_exact_ss_ed():
    _, lind = _setup_system()

    dm_ss = nk.exact.steady_state(lind, method="ed", sparse=True)
    Lop = lind.to_sparse()

    mat = np.abs(Lop @ dm_ss.reshape(-1))
    print(mat)
    assert np.all(mat == approx(0.0, rel=1e-4, abs=1e-4))
    assert dm_ss.trace() - 1 == approx(0.0, rel=1e-5, abs=1e-8)

    # dm_ss_d = nk.exact.steady_state(lind, method="ed", sparse=False)
    # Lop = lind.to_sparse()

    # mat = np.abs(dm_ss - dm_ss_d)
    # print(mat)
    # assert np.all(mat == approx(0.0, rel=1e-4, abs=1e-4))


def test_exact_ss_iterative():
    _, lind = _setup_system()

    dm_ss = nk.exact.steady_state(lind, sparse=True, method="iterative", tol=1e-5)
    Lop = lind.to_linear_operator()

    mat = np.abs(Lop @ dm_ss.reshape(-1))
    print(mat)
    assert np.all(mat == approx(0.0, rel=1e-5, abs=1e-5))
    assert dm_ss.trace() - 1 == approx(0.0, rel=1e-5, abs=1e-5)



from .abstract_optimizer import AbstractOptimizer

from .ada_delta import AdaDelta
from .ada_grad import AdaGrad
from .ada_max import AdaMax
from .ams_grad import AmsGrad
from .momentum import Momentum
from .rms_prop import RmsProp
from .sgd import Sgd
from .stochastic_reconfiguration import SR

from netket.utils import (
    jax_available as _jax_available,
    torch_available as _torch_available,
)

if _jax_available:
    from . import jax

if _torch_available:
    from .torch import Torch



import jax
import jax.numpy as jnp
from functools import partial
from netket.legacy.stats import sum_inplace as _sum_inplace
from netket.utils.mpi import n_nodes

# Stochastic Reconfiguration with jvp and vjp

# Here O (Oks) is the jacobian (derivatives w.r.t. params) of the vectorised (in x) log wavefunction (forward_fn) evaluated at all samples.
# instead of computing (and storing) the full jacobian matrix jvp and vjp are used to implement the matrix vector multiplications with it.
# Expectation values are then just the mean over the leading dimension.


def tree_conj(t):
    r"""
    conjugate all complex leaves
    The real leaves are left untouched.

    t: pytree
    """
    return jax.tree_map(lambda x: jax.lax.conj(x) if jnp.iscomplexobj(x) else x, t)


def tree_dot(a, b):
    r"""
    compute the dot product of of the flattened arrays of a and b (without actually flattening)

    a, b: pytrees with the same treedef
    """
    res = jax.tree_util.tree_reduce(
        jax.numpy.add, jax.tree_map(jax.numpy.sum, jax.tree_multimap(jax.lax.mul, a, b))
    )
    # convert shape from () to (1,)
    # this is needed for automatic broadcasting to work also when transposed with linear_transpose
    return jnp.expand_dims(res, 0)


def tree_cast(x, target):
    r"""
    Cast each leaf of x to the dtype of the corresponding leaf in target.
    The imaginary part of complex leaves which are cast to real is discarded

    x: a pytree with arrays as leaves
    target: a pytree with the same treedef as x where only the dtypes of the leaves are accessed
    """
    # astype alone would also work, however that raises ComplexWarning when casting complex to real
    # therefore the real is taken first where needed
    return jax.tree_multimap(
        lambda x, target: (x if jnp.iscomplexobj(target) else x.real).astype(
            target.dtype
        ),
        x,
        target,
    )


def tree_axpy(a, x, y):
    r"""
    compute a * x + y

    a: scalar
    x, y: pytrees with the same treedef
    """
    return jax.tree_multimap(lambda x_, y_: a * x_ + y_, x, y)


def O_jvp(x, params, v, forward_fn):
    # TODO apply the transpose of sum_inplace (allreduce) to v here
    # in order to get correct transposition with MPI
    _, res = jax.jvp(lambda p: forward_fn(p, x), (params,), (v,))
    return res


def O_vjp(x, params, v, forward_fn, return_vjp_fun=False, vjp_fun=None, allreduce=True):
    if vjp_fun is None:
        _, vjp_fun = jax.vjp(forward_fn, params, x)
    res, _ = vjp_fun(v)

    if allreduce:
        res = jax.tree_map(_sum_inplace, res)

    if return_vjp_fun:
        return res, vjp_fun
    else:
        return res


def O_mean(samples, params, forward_fn, **kwargs):
    r"""
    compute \langle O \rangle
    i.e. the mean of the rows of the jacobian of forward_fn
    """
    dtype = jax.eval_shape(forward_fn, params, samples).dtype
    v = jnp.ones(samples.shape[0], dtype=dtype) * (1.0 / (samples.shape[0] * n_nodes))
    return O_vjp(samples, params, v, forward_fn, **kwargs)


def OH_w(samples, params, w, forward_fn, **kwargs):
    r"""
    compute  O^H w
    (where ^H is the hermitian transpose)
    """
    # O^H w = (w^H O)^H
    # The transposition of the 1D arrays is omitted in the implementation:
    # (w^H O)^H -> (w* O)*

    # TODO The allreduce in O_vjp could be deferred until after the tree_cast
    # where the amount of data to be transferred would potentially be smaller
    res = tree_conj(O_vjp(samples, params, w.conjugate(), forward_fn, **kwargs))
    #
    return tree_cast(res, params)


def Odagger_DeltaO_v(samples, params, v, forward_fn, vjp_fun=None):
    r"""
    compute \langle O^\dagger \DeltaO \rangle v
    where \DeltaO = O - \langle O \rangle

    optional: pass jvp_fun to be reused
    """

    # here the allreduce is deferred until after the dot product,
    # where only scalars instead of vectors have to be summed
    # the vjp_fun is returned so that it can be reused in OH_w below
    omean, vjp_fun = O_mean(
        samples,
        params,
        forward_fn,
        return_vjp_fun=True,
        vjp_fun=vjp_fun,
        allreduce=False,
    )
    omeanv = tree_dot(omean, v)  # omeanv = omean.dot(v); is a scalar
    omeanv = _sum_inplace(omeanv)  # MPI Allreduce w/ MPI_SUM

    # v_tilde is an array of size n_samples; each MPI rank has its own slice
    v_tilde = O_jvp(samples, params, v, forward_fn)
    # v_tilde -= omeanv (elementwise):
    v_tilde = v_tilde - omeanv
    # v_tilde /= n_samples (elementwise):
    v_tilde = v_tilde * (1.0 / (samples.shape[0] * n_nodes))

    return OH_w(samples, params, v_tilde, forward_fn, vjp_fun=vjp_fun)


# TODO allow passing vjp_fun from e.g. a preceding gradient calculation with the same samples
# and optionally return vjp_fun so that it can be reused in subsequent calls
# TODO block the computations (in the same way as done with MPI) if memory consumtion becomes an issue
def mat_vec(v, forward_fn, params, samples, diag_shift):
    r"""
    compute (S + diag_shift) v

    where the elements of S are given by
    S_kl = \langle O_k^\dagger \DeltaO_ \rangle
    \DeltaO_k = O_k - \langle O_k \rangle
    and O_k (operator) is derivative of the log wavefunction w.r.t parameter k
    The expectation values are calculated as mean over the samples

    v: a pytree with the same structure as params
    forward_fn(params, x): a vectorised function returning the logarithm of the wavefunction for each configuration in x
    params: pytree of parameters with arrays as leaves
    samples: an array of samples (when using MPI each rank has its own slice of samples)
    diag_shift: a scalar diagonal shift
    """

    res = Odagger_DeltaO_v(samples, params, v, forward_fn)
    # add diagonal shift:
    res = tree_axpy(diag_shift, v, res)  # res += diag_shift * v
    return res



import abc
from .. import AbstractMachine
from . import Diagonal
import numpy as _np


class AbstractDensityMatrix(AbstractMachine):
    """Abstract class for NetKet density matrices"""

    def __init__(self, hilbert, dtype):
        super().__init__(hilbert, dtype)

    @abc.abstractmethod
    def log_val(self, xr, xc=None, out=None):
        r"""Computes the logarithm of the density matrix for a batch of visible
        quantum numbers `(xr,xc)` and stores the result into `out`.
        Specifically, for each element of the batch i, this function should compute
        out[i]=log(rho(xr[i],xc[i])).
        If xr is None, it is assumed that xr has twice as many quantum numbers and
        contains both row and columns, stored contigously.

        Args:
            xr: A matrix of `float64` of shape `(*, self.n_visible)` if xc is given.
                If xc is None, then this should be a matrix of `float64` of shape `(*, 2*self.n_visible)`.
            xc (optional): A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The length of `out` should be `xr.shape[0]`.

        Returns:
            A vector out[i]=log(rho(xr[i],xc[i])).
        """
        raise NotImplementedError

    def der_log(self, xr, xc=None, out=None):
        r"""Computes the gradient of the logarithm of the density matrix for a
        batch of visible configurations `(xr,xc)` and stores the result into `out`.

        Args:
            xr: A matrix of `float64` of shape `(*, self.input_size/2)` if xc is given.
                If xc is None, then this should be a matrix of `float64` of shape `(*, self.input_size)`.
            xc (optional): A matrix of `float64` of shape `(*, self.input_size/2)`.
            out: Destination tensor of `complex128`.
                `out` should be a matrix of shape `(x.shape[0], self.n_par)`.

        Returns:
            `out`
        """
        raise NotImplementedError

    def diagonal(self):
        r"""Returns a view of this density matrix as a Machine computing
            log_val(x)=rho(x,x).

        Returns:
            `diagonal`
        """

        return Diagonal(self)

    def to_matrix(self, normalize=True):
        r"""
        Returns a numpy array representation of the density matrix.
        Note that, in general, the size of the matrix is exponential
        in the number of quantum numbers, and this operation should thus
        only be performed for low-dimensional Hilbert spaces.

        This method requires an indexable Hilbert space.

        Args:
            normalize (bool): If True, the returned matrix is normalized in such a way that Tr(rho) =1.

        Returns:
            numpy.array: The matrix rho(x,x') for all states x,x' in the Hilbert space.

        """
        if self.hilbert.is_indexable:
            n_states = self.hilbert.n_states
            return self.to_array(normalize="trace").reshape((n_states, n_states))

        else:
            raise RuntimeError("The hilbert space is not indexable")

    def to_array(self, normalize=True):
        r"""
        Returns a numpy array representation of the density matrix.
        Note that, in general, the size of the array is exponential
        in the number of quantum numbers, and this operation should thus
        only be performed for low-dimensional Hilbert spaces.

        This method requires an indexable Hilbert space.

        Args:
            normalize (bool): If True, the returned array is normalized in such a way that Tr(rho)=1.

        Returns:
            numpy.array: The array machine(x) for all states (xr,xc) in the Hilbert space.

        """
        if self.hilbert.is_indexable:
            import itertools

            all_states = self.hilbert.all_states()
            x = _np.asarray(list(itertools.product(all_states, all_states)))

            n_states = self.hilbert.n_states

            rho_vec = self.log_val(x[:, 0, :], x[:, 1, :])

            logmax = x.real.max()
            rho_vec = _np.exp(rho_vec - logmax)

            if normalize == "trace":
                norm = _np.trace(rho_vec.reshape((n_states, n_states)))
                rho_vec /= norm
            elif normalize == True or normalize == "L2":
                norm = _np.linalg.norm(rho_vec)
                rho_vec /= norm

            return rho_vec
        else:
            raise RuntimeError("The hilbert space is not indexable")

    def log_trace(self):
        r"""
        Returns the log of the trace of the density matrix.
        Note that, in general, the size of the density matrix is exponential
        in the number of quantum numbers, and this operation should thus
        only be performed for low-dimensional Hilbert spaces.

        This method requires an indexable Hilbert space.

        Returns:
            float: log(Tr(rho))

        """
        if self.hilbert.is_indexable:
            all_states = self.hilbert.all_states()
            rho_diag = self.log_val(all_states, all_states).real
            maxl = rho_diag.max()
            log_t = _np.log(_np.exp((rho_diag - maxl).sum()))

            return log_t + maxl
        else:
            raise RuntimeError("The hilbert space is not indexable")

    @property
    def input_size(self):
        return 2 * self.hilbert.size



from ..abstract_optimizer import AbstractOptimizer
import numpy as _np


class AmsGrad(AbstractOptimizer):
    r"""AmsGrad Optimizer.
        In some cases, adaptive learning rate methods such as AdaMax fail
        to converge to the optimal solution because of the exponential
        moving average over past gradients. To address this problem,
        Sashank J. Reddi, Satyen Kale and Sanjiv Kumar proposed the
        AmsGrad [update algorithm](https://openreview.net/forum?id=ryQu7f-RZ).
        The update rule for :math:`\mathbf{v}` (equivalent to :math:`E[g^2]` in AdaDelta
        and :math:`\mathbf{s}` in RMSProp) is modified such that :math:`v^\prime_k \geq v_k`
        is guaranteed, giving the algorithm a "long-term memory" of past gradients.
        The vectors :math:`\mathbf{m}` and :math:`\mathbf{v}` are initialized to zero, and
        are updated with the parameters :math:`\mathbf{p}`:

        .. math:: m^\prime_k &= \beta_1 m_k + (1-\beta_1)G_k(\mathbf{p})\\
                  v^\prime_k &= \beta_2 v_k + (1-\beta_2)G_k(\mathbf{p})^2\\
                  v^\prime_k &= \mathrm{Max}(v^\prime_k, v_k)\\
                  p^\prime_k &= p_k - \frac{\eta}{\sqrt{v^\prime_k}+\epsilon}m^\prime_k

    """

    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, epscut=1.0e-7):
        r"""
        Constructs a new ``AmsGrad`` optimizer.

        Args:
            learning_rate: The learning rate $\eta$.
            beta1: First exponential decay rate.
            beta2: Second exponential decay rate.
            epscut: Small epsilon cutoff.

        Examples:
            Simple AmsGrad optimizer.

            >>> from netket.optimizer import AmsGrad
            >>> op = AmsGrad()
        """

        if epscut <= 0:
            raise ValueError("Invalid epsilon cutoff.")
        if learning_rate < 0:
            raise ValueError("Invalid learning rate.")
        if beta1 < 0 or beta1 > 1:
            raise ValueError("Invalid beta1.")
        if beta2 < 0 or beta2 > 1:
            raise ValueError("Invalid beta1.")

        self._eta = learning_rate
        self._beta1 = beta1
        self._beta2 = beta2
        self._epscut = epscut

        self._mt = None
        self._vt = None
        self._niter = 0

    def update(self, grad, pars):
        if self._mt is None:
            self._mt = _np.zeros(pars.shape[0])
            self._vt = _np.zeros(pars.shape[0])

        self._mt = self._beta1 * self._mt + (1.0 - self._beta1) * grad

        self._vt = _np.maximum(
            self._vt, self._beta2 * self._vt + (1 - self._beta2) * _np.abs(grad) ** 2.0
        )

        pars -= self._eta * self._mt / _np.sqrt(self._vt + self._epscut)

        return pars

    def reset(self):
        if self._mt is not None:
            self._mt.fill(0.0)
            self._vt.fill(0.0)

    def __repr__(self):
        rep = "AmsGrad optimizer with these parameters :"
        rep += "\nLearning rate = " + str(self._eta)
        rep += "\nbeta1 = " + str(self._beta1)
        rep += "\nbeta2 = " + str(self._beta2)
        rep += "\nepscut = " + str(self._epscut)
        return rep



import math
from typing import Any, Optional, Tuple, List

from netket import legacy as _nk

from netket.operator import (
    local_values as _local_values,
    der_local_values as _der_local_values,
)
from netket.legacy.stats import (
    statistics as _statistics,
    mean as _mean,
    sum_inplace as _sum_inplace,
)

from netket.legacy.vmc_common import info, tree_map, trees2_map
from netket.legacy.abstract_variational_driver import AbstractVariationalDriver
import operator


class SteadyState(AbstractVariationalDriver):
    """
    Variational steady-state search by minimization of the L2,2 norm of L\rho using
    Variational Monte Carlo (VMC).
    """

    def __init__(
        self,
        lindblad,
        sampler,
        optimizer,
        n_samples,
        n_discard=None,
        sr=None,
        sampler_obs=None,
        n_samples_obs=None,
        n_discard_obs=None,
    ):
        """
        Initializes the driver class.

        Args:
            liouvillian: The liouvillian of the system.
            sampler: The Monte Carlo sampler for the density matrix.
            optimizer: Determines how optimization steps are performed given the
                bare energy gradient. This parameter supports three different kinds of inputs,
                which are described in the docs of `make_optimizer_fn`.
            n_samples: Number of Markov Chain Monte Carlo sweeps to be
                performed at each step of the optimization.
            n_discard (int, optional): Number of sweeps to be discarded at the
                beginning of the sampling, at each step of the optimization.
                Defaults to 10% of the number of samples allocated to each MPI node.
            sr (SR, optional): Determines whether and how stochastic reconfiguration
                is applied to the bare energy gradient before performing applying
                the optimizer. If this parameter is not passed or None, SR is not used.
            sampler_obs: The Monte Carlo sampler for the diagonal of the density matrix, used
                to compute observables (default: same as sampler).
            n_samples_obs: n_samples for the observables (default: n_samples)
            n_discard_obs: n_discard for the observables (default: n_discard)

        """
        super(SteadyState, self).__init__(
            sampler.machine, optimizer, minimized_quantity_name="LdagL"
        )  #'\u3008L\u2020L\u3009')

        self._lind = lindblad
        self._sampler = sampler
        self._sampler_obs = sampler_obs

        self.sr = sr

        self._npar = self._machine.n_par

        self._batch_size = sampler.sample_shape[0]

        # Check how many parallel nodes we are running on
        self.n_nodes = _nk.utils.mpi.n_nodes

        self.n_samples = n_samples
        self.n_discard = n_discard

        self._obs_samples_valid = False
        if sampler_obs is not None:
            # Set the machine_pow of the sampler over the diagonal of the density matrix
            # to be |\rho(x,x)|
            sampler_obs.machine_pow = 1.0
            self._batch_size_obs = sampler_obs.sample_shape[0]
            self.n_samples_obs = n_samples_obs
            self.n_discard_obs = n_discard_obs

        self._der_logs_ave = None
        self._lloc = None

        self._dp = None

    @property
    def sr(self):
        return self._sr

    @sr.setter
    def sr(self, sr):
        self._sr = sr
        if self._sr is not None:
            self._sr.setup(self.machine)

    @property
    def n_samples(self):
        return self._n_samples

    @property
    def ldagl(self):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.
        """
        return self._loss_stats

    @property
    def n_samples_obs(self):
        return self._n_samples_obs

    @n_samples.setter
    def n_samples(self, n_samples):
        if n_samples <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples={}".format(n_samples)
            )
        n_samples_chain = int(math.ceil((n_samples / self._batch_size)))
        self._n_samples_node = int(math.ceil(n_samples_chain / self.n_nodes))

        self._n_samples = int(self._n_samples_node * self._batch_size * self.n_nodes)

        self._samples = None
        self._grads = None
        self._jac = None

        self._der_logs = None
        self._der_loc_vals = None

    @n_samples_obs.setter
    def n_samples_obs(self, n_samples):
        if self._sampler_obs is None:
            return

        if n_samples <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples_obs={}".format(n_samples)
            )

        n_samples_chain = int(math.ceil((n_samples / self._batch_size_obs)))
        self._n_samples_node_obs = int(math.ceil(n_samples_chain / self.n_nodes))

        self._n_samples_obs = int(
            self._n_samples_node_obs * self._batch_size_obs * self.n_nodes
        )

        self._samples_obs = None

    @property
    def n_discard(self):
        return self._n_discard

    @property
    def n_discard_obs(self):
        return self._n_discard_obs

    @n_discard.setter
    def n_discard(self, n_discard: Optional[int]):
        if n_discard is not None and n_discard < 0:
            raise ValueError(
                "Invalid number of discarded samples: n_discard={}".format(n_discard)
            )
        self._n_discard = (
            int(n_discard)
            if n_discard != None
            else self._n_samples_node * self._batch_size // 10
        )

    @n_discard_obs.setter
    def n_discard_obs(self, n_discard: Optional[int]):
        if self._sampler_obs is None:
            return

        if n_discard is not None and n_discard < 0:
            raise ValueError(
                "Invalid number of discarded samples: n_discard_obs={}".format(
                    n_discard
                )
            )
        self._n_discard_obs = (
            int(n_discard)
            if n_discard is not None
            else self._n_samples_node_obs * self._batch_size_obs // 10
        )

    def _forward_and_backward(self, sample=True):
        """
        Performs a number of VMC optimization steps.

        Args:
            n_steps (int): Number of steps to perform.
        """

        if sample:
            self._sampler.reset()
            self._obs_samples_valid = False

            # Burnout phase
            self._sampler.generate_samples(self._n_discard)

            # Generate samples and store them
            self._samples = self._sampler.generate_samples(
                self._n_samples_node, samples=self._samples
            )

        # Estimate C^[loc] (local energy) and LdagL
        self._lloc, self._loss_stats = self._get_mc_superop_stats(self._lind)

        # Flatten chain dimension
        samples_r = self._samples.reshape((-1, self._samples.shape[-1]))
        lloc_r = self._lloc.reshape(-1, 1)

        # Compute Log derivatives
        self._der_logs = self._machine.der_log(samples_r)
        # Compute statistical average (also across nodes)
        self._der_logs_ave = tree_map(_mean, self._der_logs, axis=0)

        # Compute \nabla C^[Loc]
        self._der_loc_vals = _der_local_values(
            self._lind, self._machine, samples_r, center_derivative=False
        )

        # apply this function to every element of the gradient.
        n_samples_node = lloc_r.shape[0]

        def gradfun(der_loc_vals, der_logs_ave):
            par_dims = der_loc_vals.ndim - 1

            _lloc_r = lloc_r.reshape(
                (n_samples_node,) + tuple(1 for i in range(par_dims))
            )

            grad = _mean(der_loc_vals.conjugate() * _lloc_r, axis=0) - (
                der_logs_ave.conjugate() * self._loss_stats.mean
            )
            return grad

        self._grads = trees2_map(gradfun, self._der_loc_vals, self._der_logs_ave)

        # Perform update
        if self._sr:
            # Center the log derivatives
            self._jac = trees2_map(
                lambda x, y: x - y, self._der_logs, self._der_logs_ave
            )

            self._dp = self._sr.compute_update(self._jac, self._grads, self._dp)

        else:
            #  if Real pars but complex gradient, take only real part
            # not necessary for SR because sr already does it.
            if not self._machine.has_complex_parameters:
                self._dp = tree_map(lambda x: x.real, self._grads)
            else:
                self._dp = self._grads

        return self._dp

    def sweep_diagonal(self):
        """
        Sweeps the diagonal of the density matrix with the observable sampler.
        """
        self._sampler_obs.reset()

        # Burnout phase
        self._sampler_obs.generate_samples(self._n_discard)

        # Generate samples and store them
        self._samples_obs = self._sampler_obs.generate_samples(
            self._n_samples_node_obs, samples=self._samples_obs
        )

        self._obs_samples_valid = True

    def _estimate_stats(self, obs):
        return self._get_mc_obs_stats(obs)[1]

    def reset(self):
        self._sampler.reset()
        super().reset()

    def _get_mc_superop_stats(self, op):
        samples_r = self._samples.reshape((-1, self._samples.shape[-1]))

        loc = _local_values(op, self._machine, samples_r).reshape(
            self._samples.shape[0:2]
        )

        # notice that loc.T is passed to statistics, since that function assumes
        # that the first index is the batch index.
        return loc, _statistics(abs(loc.T) ** 2)

    def _get_mc_obs_stats(self, op):
        if not self._obs_samples_valid:
            self.sweep_diagonal()

        samples_r = self._samples_obs.reshape((-1, self._samples_obs.shape[-1]))

        loc = _local_values(op, self._machine, samples_r).reshape(
            self._samples_obs.shape[0:2]
        )

        # notice that loc.T is passed to statistics, since that function assumes
        # that the first index is the batch index.
        return loc, _statistics(loc.T)

    def __repr__(self):
        return "SteadyState(step_count={}, n_samples={}, n_discard={})".format(
            self.step_count, self.n_samples, self.n_discard
        )

    def info(self, depth=0):
        lines = [
            "{}: {}".format(name, info(obj, depth=depth + 1))
            for name, obj in [
                ("Liouvillian", self._lind),
                ("Machine", self._machine),
                ("Optimizer", self._optimizer),
                ("SR solver", self._sr),
            ]
        ]
        return "\n  ".join([str(self)] + lines)



# Copyright 2018-2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket import legacy as nk

# 1D Lattice
L = 20
g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
# with total Sz equal to 0
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes, total_sz=0)

# Heisenberg hamiltonian
ha = nk.operator.Heisenberg(hilbert=hi)


# Fully-connected machine
layers = (
    nk.layer.FullyConnected(input_size=L, output_size=(2 * L), use_bias=True),
    nk.layer.Lncosh(input_size=(2 * L)),
    nk.layer.SumOutput(input_size=(2 * L)),
)
for layer in layers:
    layer.init_random_parameters(seed=12345, sigma=0.01)

ffnn = nk.machine.FFNN(hi, layers)

sa = nk.sampler.MetropolisExchange(machine=ffnn)


# Optimizer
op = nk.optimizer.Sgd(ma, learning_rate=0.05)

# Stochastic reconfiguration
gs = nk.variational.Vmc(
    hamiltonian=ha,
    sampler=sa,
    optimizer=op,
    n_samples=1000,
    diag_shift=0.1,
    method="Sr",
    use_iterative=True,
)

gs.run(out="test", n_iter=300)



import abc
import numbers

import netket as _nk
import numpy as _np

from netket.legacy.logging import JsonLog as _JsonLog

from netket.legacy.vmc_common import tree_map

from netket.utils.mpi import node_number as _rank, n_nodes as _n_nodes

from tqdm import tqdm

import warnings


def _to_iterable(maybe_iterable):
    """
    _to_iterable(maybe_iterable)

    Ensure the result is iterable. If the input is not iterable, it is wrapped into a tuple.
    """
    if hasattr(maybe_iterable, "__iter__"):
        surely_iterable = maybe_iterable
    else:
        surely_iterable = (maybe_iterable,)

    return surely_iterable


# Note: to implement a new Driver (see also _vmc.py for an example)
# If you want to inherit the nice interface of AbstractMCDriver, you should
# subclass it, defining the following methods:
# - Either _forward_and_backward or individually _forward, _backward, that should
#   compute the loss function and the gradient. If the driver is minimizing or
#   maximising some loss function, this quantity should be assigned to self._stats
#   in order to monitor it.
# - _estimate_stats should return the MC estimate of a single operator
# - reset should reset the driver (usually the sampler).
# - info should return a string with an overview of the driver.
# - The __init__ method shouldbe called with the machine and the optimizer. If this
#   driver is minimising a loss function and you want it's name to show up automatically
#   in the progress bar/ouput files you should pass the optional keyword argument
#   minimized_quantity_name.
class AbstractVariationalDriver(abc.ABC):
    """Abstract base class for NetKet Variational Monte Carlo drivers"""

    def __init__(self, machine, optimizer, minimized_quantity_name=""):
        self._mynode = _rank
        self._mpi_nodes = _n_nodes
        self._loss_stats = None
        self._loss_name = minimized_quantity_name
        self._step_count = 0

        self._machine = machine
        self._optimizer = optimizer

    def _forward_and_backward(self):
        """
        Performs the forward and backward pass at the same time.
        Concrete drivers should either override this method, or override individually
        _forward and _backward.

        :return: the update for the weights.
        """
        self._forward()
        dp = self._backward()
        return dp

    def _forward(self):
        """
        Performs the forward pass, computing the loss function.
        Concrete should either implement _forward and _backward or the joint method
        _forward_and_backward.
        """
        raise NotImplementedError()

    def _backward(self):
        """
        Performs the backward pass, computing the update for the parameters.
        Concrete should either implement _forward and _backward or the joint method
        _forward_and_backward.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _estimate_stats(self, observable):
        """
        Returns the MCMC statistics for the expectation value of an observable.
        Must be implemented by super-classes of AbstractVMC.

        :param observable: A quantum operator (netket observable)
        :return:
        """
        pass

    @abc.abstractmethod
    def reset(self):
        """
        Resets the driver.
        Concrete drivers should also call super().reset() to ensure that the step
        count is set to 0.
        """
        self.step_count = 0
        pass

    @abc.abstractmethod
    def info(self, depth=0):
        """
        Returns an info string used to print information to screen about this driver.
        """
        pass

    @property
    def machine(self):
        """
        Returns the machine that is optimized by this driver.
        """
        return self._machine

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer_new):
        self._optimizer = optimizer_new

    @property
    def step_count(self):
        """
        Returns a monotonic integer labelling all the steps performed by this driver.
        This can be used, for example, to identify the line in a log file.
        """
        return self._step_count

    def iter(self, n_steps, step=1):
        """
        Returns a generator which advances the VMC optimization, yielding
        after every `step_size` steps.

        Args:
            :n_iter (int=None): The total number of steps to perform.
            :step_size (int=1): The number of internal steps the simulation
                is advanced every turn.

        Yields:
            int: The current step.
        """
        for _ in range(0, n_steps, step):
            for i in range(0, step):
                dp = self._forward_and_backward()
                if i == 0:
                    yield self.step_count

                self.update_parameters(dp)

    def advance(self, steps=1):
        """
        Performs `steps` optimization steps.

        :param steps: (Default=1) number of steps
        """
        for _ in self.iter(steps):
            pass

    def run(
        self,
        n_iter,
        out=None,
        obs=None,
        show_progress=True,
        save_params_every=50,  # for default logger
        write_every=50,  # for default logger
        step_size=1,  # for default logger
        callback=lambda *x: True,
    ):
        """
        Executes the Monte Carlo Variational optimization, updating the weights of the network
        stored in this driver for `n_iter` steps and dumping values of the observables `obs`
        in the output `logger`. If no logger is specified, creates a json file at `out`,
        overwriting files with the same prefix.

        Args:
            :n_iter: the total number of iterations
            :out: A logger object, or an iterable of loggers, to be used to store simulation log and data.
                If this argument is a string, it will be used as output prefix for the standard JSON logger.
            :obs: An iterable containing all observables that should be computed
            :save_params_every: Every how many steps the parameters of the network should be
            serialized to disk (ignored if logger is provided)
            :write_every: Every how many steps the json data should be flushed to disk (ignored if
            logger is provided)
            :step_size: Every how many steps should observables be logged to disk (default=1)
            :show_progress: If true displays a progress bar (default=True)
            :callback: Callable or list of callable callback functions to stop training given a condition
        """

        if not isinstance(n_iter, numbers.Number):
            raise ValueError(
                "n_iter, the first positional argument to `run`, must be a number!"
            )

        if obs is None:
            obs = {}

        if out is None:
            out = tuple()
            print(
                "No output specified (out=[apath|nk.logging.JsonLogger(...)])."
                "Running the optimization but not saving the output."
            )

        # Log only non-root nodes
        if self._mynode == 0:
            # if out is a path, create an overwriting Json Log for output
            if isinstance(out, str):
                loggers = (_JsonLog(out, "w", save_params_every, write_every),)
            else:
                loggers = _to_iterable(out)
        else:
            loggers = tuple()
            show_progress = False

        callbacks = _to_iterable(callback)
        callback_stop = False

        with tqdm(
            self.iter(n_iter, step_size), total=n_iter, disable=not show_progress
        ) as itr:
            for step in itr:

                log_data = self.estimate(obs)

                # if the cost-function is defined then report it in the progress bar
                if self._loss_stats is not None:
                    itr.set_postfix_str(self._loss_name + "=" + str(self._loss_stats))
                    log_data[self._loss_name] = self._loss_stats

                for logger in loggers:
                    logger(self.step_count, log_data, self.machine)

                for callback in callbacks:
                    if not callback(step, log_data, self):
                        callback_stop = True

                if callback_stop:
                    break

        # flush at the end of the evolution so that final values are saved to
        # file
        for logger in loggers:
            logger.flush(self.machine)

    def estimate(self, observables):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.

        Args:
            observables: A pytree of operators for which statistics should be computed.

        Returns:
            A pytree of the same structure as the input, containing MCMC statistics
            for the corresponding operators as leaves.
        """
        return tree_map(self._estimate_stats, observables)

    def update_parameters(self, dp):
        """
        Updates the parameters of the machine using the optimizer in this driver

        Args:
            :param dp: the gradient
        """
        self._machine.parameters = self._optimizer.update(dp, self._machine.parameters)
        self._step_count += 1



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import numpy as np

from netket.utils.dispatch import parametric

from .abstract_hilbert import AbstractHilbert
from .discrete_hilbert import DiscreteHilbert


@parametric
class DoubledHilbert(DiscreteHilbert):
    r"""
    Superoperatorial hilbert space for states living in the tensorised state
    :math:`\hat{H}\otimes \hat{H}`, encoded according to Choi's isomorphism.
    """

    def __init__(self, hilb: AbstractHilbert):
        r"""
        Superoperatorial hilbert space for states living in the tensorised
        state :math:`\hat{H}\otimes \hat{H}`, encoded according to Choi's
        isomorphism.

        Args:
            hilb: the Hilbert space H.

        Examples:
            Simple superoperatorial hilbert space for few spins.

            >>> import netket as nk
            >>> g = nk.graph.Hypercube(length=5,n_dim=2,pbc=True)
            >>> hi = nk.hilbert.Spin(N=3, s=0.5)
            >>> hi2 = nk.hilbert.DoubledHilbert(hi)
            >>> print(hi2.size)
            6
        """
        self.physical = hilb
        self._size = 2 * hilb.size

        super().__init__(shape=hilb.shape * 2)

    @property
    def size(self):
        return self._size

    @property
    def shape(self):
        return self._shape

    @property
    def is_finite(self):
        return self.physical.is_finite

    @property
    def local_size(self):
        return self.physical.local_size

    @property
    def local_states(self):
        return self.physical.local_states

    def size_at_index(self, i: int) -> int:
        return self.physical.size_at_index(
            i if i < self.physical.size else i - self.physical.size
        )

    def states_at_index(self, i: int) -> Optional[List[float]]:
        r"""A list of discrete local quantum numbers at the site i.
        If the local states are infinitely many, None is returned."""
        return self.physical.states_at_index(
            i if i < self.physical.size else i - self.physical.size
        )

    @property
    def size_physical(self):
        return self.physical.size

    @property
    def n_states(self):
        return self.physical.n_states ** 2

    def _numbers_to_states(self, numbers, out):
        # !!! WARNING
        # This code assumes that states are stored in a MSB
        # (Most Significant Bit) format.
        # We assume that the rightmost-half indexes the LSBs
        # and the leftmost-half indexes the MSBs
        # HilbertIndex-generated states respect this, as they are:
        # 0 -> [0,0,0,0]
        # 1 -> [0,0,0,1]
        # 2 -> [0,0,1,0]
        # etc...

        n = self.physical.size
        dim = self.physical.n_states
        left, right = np.divmod(numbers, dim)

        self.physical.numbers_to_states(left, out=out[:, 0:n])
        self.physical.numbers_to_states(right, out=out[:, n : 2 * n])

        return out

    def _states_to_numbers(self, states, out):
        # !!! WARNING
        # See note above in numbers_to_states

        n = self.physical.size
        dim = self.physical.n_states

        self.physical._states_to_numbers(states[:, 0:n], out=out)
        _out_l = out * dim

        self.physical._states_to_numbers(states[:, n : 2 * n], out=out)
        out += _out_l

        return out

    def __repr__(self):
        return "DoubledHilbert({})".format(self.physical)

    @property
    def _attrs(self):
        return (self.physical,)

    def _random_state_legacy(self, size=None, *, out=None, rgen=None):
        if isinstance(size, int):
            size = (size,)
        shape = (*size, self.size) if size is not None else (self.size,)

        if out is None:
            out = np.empty(shape=shape)

        n = self.size_physical

        self.physical.random_state(out=out[..., :n], size=size, rgen=rgen)
        self.physical.random_state(out=out[..., n:], size=size, rgen=rgen)

        return out



# Copyright 2021 The NetKet Authors - All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
import numpy as np

# 1D Lattice
L = 64

g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)

# Ising spin hamiltonian
ha = nk.operator.Ising(hilbert=hi, graph=g, h=0.0)

# RBM Spin Machine
ma = nk.models.Jastrow(dtype=np.float64)

# Metropolis Local Sampling
sa = nk.sampler.MetropolisLocal(hi, n_chains=16)

# Optimizer
op = nk.optimizer.Sgd(learning_rate=0.05)
sr = nk.optimizer.SR(diag_shift=0.01)

# Variational monte carlo driver
gs = nk.VMC(ha, op, sa, ma, n_samples=8000, sr=sr)

# Run the optimization for 300 iterations
gs.run(
    n_iter=300,
    out="test",
    # stop if variance is essentially zero (= reached eigenstate)
    callback=nk.callbacks.EarlyStopping(
        monitor="variance", baseline=1e-12, patience=np.infty
    ),
)



from functools import singledispatch
from . import numpy


@singledispatch
def AdaGrad(machine, learning_rate=0.001, epscut=1.0e-7):
    r"""AdaGrad Optimizer.
        In many cases, in Sgd the learning rate :math`\eta` should
        decay as a function of training iteration to prevent overshooting
        as the optimum is approached. AdaGrad is an adaptive learning
        rate algorithm that automatically scales the learning rate with a sum
        over past gradients. The vector :math:`\mathbf{g}` is initialized to zero.
        Given a stochastic estimate of the gradient of the cost function :math:`G(\mathbf{p})`,
        the updates for :math:`g_k` and the parameter :math:`p_k` are


        .. math:: g^\prime_k &= g_k + G_k(\mathbf{p})^2\\
                  p^\prime_k &= p_k - \frac{\eta}{\sqrt{g_k + \epsilon}}G_k(\mathbf{p})

        AdaGrad has been shown to perform particularly well when
        the gradients are sparse, but the learning rate may become too small
        after many updates because the sum over the squares of past gradients is cumulative.


        Args:
           learning_rate: Learning rate :math:`\eta`.
           epscut: Small :math:`\epsilon` cutoff.

        Examples:
           Simple AdaGrad optimizer.

           >>> from netket.optimizer import AdaGrad
           >>> op = AdaGrad()
        """

    return numpy.AdaGrad(learning_rate, epscut)



from .metropolis_hastings import *
from ._kernels import _ExchangeKernel
import numpy as _np


def _compute_clusters(graph, d_max):
    clusters = []
    distances = _np.asarray(graph.distances())
    size = distances.shape[0]
    for i in range(size):
        for j in range(i + 1, size):
            if distances[i][j] <= d_max:
                clusters.append((i, j))

    res_clusters = _np.empty((len(clusters), 2), dtype=_np.int64)

    for i, cluster in enumerate(clusters):
        res_clusters[i] = _np.asarray(cluster)

    return res_clusters


def MetropolisExchange(
    machine, clusters=None, graph=None, d_max=1, n_chains=16, sweep_size=None, **kwargs
):
    r"""
    This sampler acts locally only on two local degree of freedom :math:`s_i` and :math:`s_j`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s^\prime_j \dots s_N`,
    where in general :math:`s^\prime_i \neq s_i` and :math:`s^\prime_j \neq s_j`.
    The sites :math:`i` and :math:`j` are also chosen to be within a maximum graph
    distance of :math:`d_{\mathrm{max}}`.

    The transition probability associated to this sampler can
    be decomposed into two steps:

    1. A pair of indices :math:`i,j = 1\dots N`, and such
       that :math:`\mathrm{dist}(i,j) \leq d_{\mathrm{max}}`,
       is chosen with uniform probability.
    2. The sites are exchanged, i.e. :math:`s^\prime_i = s_j` and :math:`s^\prime_j = s_i`.

    Notice that this sampling method generates random permutations of the quantum
    numbers, thus global quantities such as the sum of the local quantum numbers
    are conserved during the sampling.
    This scheme should be used then only when sampling in a
    region where :math:`\sum_i s_i = \mathrm{constant}` is needed,
    otherwise the sampling would be strongly not ergodic.

    Args:
          machine: A machine :math:`\Psi(s)` used for the sampling.
                   The probability distribution being sampled
                   from is :math:`F(\Psi(s))`, where the function
                   :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.

          d_max: The maximum graph distance allowed for exchanges.
          n_chains: The number of Markov Chain to be run in parallel on a single process.
          sweep_size: The number of exchanges that compose a single sweep.
                      If None, sweep_size is equal to the number of degrees of freedom (n_visible).


    Examples:
          Sampling from a RBM machine in a 1D lattice of spin 1/2, using
          nearest-neighbours exchanges.

          >>> import netket as nk
          >>>
          >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
          >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
          >>>
          >>> # RBM Spin Machine
          >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
          >>>
          >>> # Construct a MetropolisExchange Sampler
          >>> sa = nk.sampler.MetropolisExchange(machine=ma)
          >>> print(sa.machine.hilbert.size)
          100
    """
    if clusters is None and graph is not None:
        assert (
            graph.n_nodes == machine.hilbert.size
        ), "The size of the graph must match the hilbert space of the machine."
        clusters = _compute_clusters(graph, d_max)
    elif not (clusters is not None and graph is None):
        raise ValueError(
            """You must either provide the list of exchange-clusters or a netket graph, from
                          which clusters will be computed using the maximum distance d_max. """
        )

    transition_kernel = _ExchangeKernel(machine, clusters)

    return MetropolisHastings(
        machine, transition_kernel, n_chains, sweep_size, **kwargs
    )


def MetropolisExchangePt(
    machine,
    clusters=None,
    graph=None,
    d_max=1,
    n_replicas=16,
    sweep_size=None,
    **kwargs,
):
    r"""
    This sampler performs parallel-tempering
    moves in addition to the local moves implemented in `MetropolisExchange`.
    The number of replicas can be chosen by the user.

    Args:
        machine: A machine :math:`\Psi(s)` used for the sampling.
                 The probability distribution being sampled
                 from is :math:`F(\Psi(s))`, where the function
                 :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
        d_max: The maximum graph distance allowed for exchanges.
        n_replicas: The number of replicas used for parallel tempering.
        sweep_size: The number of exchanges that compose a single sweep.
                    If None, sweep_size is equal to the number of degrees of freedom (n_visible).

    Examples:
        Sampling from a RBM machine in a 1D lattice of spin 1/2, using
        nearest-neighbours exchanges.

        >>> import netket as nk
        >>>
        >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
        >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
        >>>
        >>> # RBM Spin Machine
        >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
        >>>
        >>> # Construct a MetropolisExchange Sampler with parallel tempering
        >>> sa = nk.sampler.MetropolisExchangePt(machine=ma,n_replicas=24)
        >>> print(sa.machine.hilbert.size)
        100
    """
    if clusters is None and graph is not None:
        assert (
            graph.n_nodes == machine.hilbert.size
        ), "The size of the graph must match the hilbert space of the machine."

        clusters = _compute_clusters(graph, d_max)
    elif not (clusters is not None and graph is None):
        raise ValueError(
            """You must either provide the list of exchange-clusters or a netket graph, from
                          which clusters will be computed using the maximum distance d_max. """
        )

    transition_kernel = _ExchangeKernel(machine, clusters)

    return MetropolisHastingsPt(
        machine, transition_kernel, n_replicas, sweep_size, **kwargs
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import jit

from netket.legacy.machine.density_matrix.abstract_density_matrix import (
    AbstractDensityMatrix as DensityMatrix,
)

from ._local_liouvillian import LocalLiouvillian as _LocalLiouvillian


@jit(nopython=True)
def _local_values_kernel(log_vals, log_val_primes, mels, sections, out):
    low_range = 0
    for i, s in enumerate(sections):
        out[i] = (
            mels[low_range:s] * np.exp(log_val_primes[low_range:s] - log_vals[i])
        ).sum()
        low_range = s


def _local_values_impl(op, machine, v, log_vals, out):

    sections = np.empty(v.shape[0], dtype=np.int32)
    v_primes, mels = op.get_conn_flattened(np.asarray(v), sections)

    log_val_primes = machine.log_val(v_primes)

    _local_values_kernel(
        np.asarray(log_vals), np.asarray(log_val_primes), mels, sections, out
    )


@jit(nopython=True)
def _op_op_unpack_kernel(v, sections, vold):

    low_range = 0
    for i, s in enumerate(sections):
        vold[low_range:s] = v[i]
        low_range = s

    return vold


def _local_values_op_op_impl(op, machine, v, log_vals, out):

    sections = np.empty(v.shape[0], dtype=np.int32)
    v_np = np.asarray(v)
    v_primes, mels = op.get_conn_flattened(v_np, sections)

    vold = np.empty((sections[-1], v.shape[1]))
    _op_op_unpack_kernel(v_np, sections, vold)

    log_val_primes = machine.log_val(v_primes, vold)

    _local_values_kernel(
        np.asarray(log_vals), np.asarray(log_val_primes), mels, sections, out
    )


def local_values(op, machine, v, log_vals=None, out=None):
    r"""
    Computes local values of the operator `op` for all `samples`.

    The local value is defined as
    .. math:: O_{\mathrm{loc}}(x) = \langle x | O | \Psi \rangle / \langle x | \Psi \rangle


            Args:
                op: Hermitian operator.
                v: A numpy array or matrix containing either a batch of visible
                    configurations :math:`V = v_1,\dots v_M`.
                    Each row of the matrix corresponds to a visible configuration.
                machine: Wavefunction :math:`\Psi`.
                log_vals: A scalar/numpy array containing the value(s) :math:`\Psi(V)`.
                    If not given, it is computed from scratch.
                    Defaults to None.
                out: A scalar or a numpy array of local values of the operator.
                    If not given, it is allocated from scratch and then returned.
                    Defaults to None.

            Returns:
                If samples is given in batches, a numpy array of local values
                of the operator, otherwise a scalar.
    """

    # True when this is the local_value of a densitymatrix times an operator (observable)
    is_op_times_op = isinstance(machine, DensityMatrix) and not isinstance(
        op, _LocalLiouvillian
    )
    if v.ndim != 2:
        raise RuntimeError("Invalid input shape, expected a 2d array")

    if log_vals is None:
        if not is_op_times_op:
            log_vals = machine.log_val(v)
        else:
            log_vals = machine.log_val(v, v)

    if not is_op_times_op:
        _impl = _local_values_impl
    else:
        _impl = _local_values_op_op_impl

    assert (
        v.shape[1] == op.hilbert.size
    ), "samples has wrong shape: {}; expected (?, {})".format(v.shape, op.hilbert.size)

    if out is None:
        out = np.empty(v.shape[0], dtype=np.complex128)

    _impl(op, machine, v, log_vals, out)

    return out
